"""Methods for setting and validating."""
from taurus.entity.valid_list import ValidList


def validate_list(obj, typ, trigger=None):
    """
    Attempts to return obj as a list, each element of which has type typ.

    Parameters
    ----------
    obj: Any
        Object or list of objects, all of which should be of type typ.
    typ: Type
        The desired type of obj, or if obj is a list, every element of obj.
    trigger: function
        A function to invoke when putting obj into a list.

    Returns
    -------
    ValidList
        A list with all elements of obj.
        The list is constrained so that every element has type typ.

    """
    if obj is None:
        return ValidList([], typ, trigger)
    elif isinstance(obj, list):
        return ValidList(obj, typ, trigger)
    else:
        return ValidList([obj], typ, trigger)


def validate_str(obj):
    """
    Check that obj is a string and then convert it to unicode.

    Parameters
    ----------
    obj: Any
        Object that should be a string

    Returns
    -------
    str
        `obj` as a string.

    Raises
    -------
    ValueError
        If `obj` is not a string.

    """
    if not isinstance(obj, str):
        raise ValueError("Expected a string but got {} instead".format(type(obj)))

    # If python 2 and the string isn't already unicode, turn it into unicode"""
    try:
        return obj.decode("utf-8")
    except AttributeError:
        return obj



from taurus.entity.dict_serializable import DictSerializable
from taurus.entity.template.attribute_template import AttributeTemplate
from taurus.entity.value.base_value import BaseValue
from taurus.enumeration import Origin
from taurus.entity.setters import validate_list
from taurus.entity.file_link import FileLink
from taurus.entity.link_by_uid import LinkByUID


class BaseAttribute(DictSerializable):
    """
    Base class for all attributes, which include property, condition, parameter, and metadata.

    Parameters
    ----------
    name: str
        Required name of the attribute. Each attribute within an object must have a unique name.
    notes: str
        Optional free-form notes about the attribute.
    value: :py:class:`BaseValue <taurus.entity.value.base_value.BaseValue>`
        The value of the attribute.
    template: :class:`AttributeTemplate \
    <taurus.entity.template.attribute_template.AttributeTemplate>`
        Attribute template that defines the allowed bounds of this attribute. If a template
        and value are both present then the value must be within the template bounds.
    origin: str
        The origin of the attribute. Must be one of "measured", "predicted", "summary",
        "specified", "computed", or "unknown." Default is "unknown."
    file_links: List[FileLink]
        Links to files associated with the attribute.

    """

    attribute_type = None

    def __init__(self, name=None, template=None, origin="unknown", value=None, notes=None,
                 file_links=None):
        if name is None:
            raise ValueError("Attributes must be named")
        self.name = name
        self.notes = notes

        self._value = None
        self._template = None
        self._origin = None
        self._file_links = None,

        self.value = value
        self.template = template
        self.origin = origin
        self.file_links = file_links

        # TODO: Make template required (need to fix a bunch of unit tests)
        #  if self.template is None:
        #    raise ValueError("Template is required")

    @property
    def value(self):
        """Get value."""
        return self._value

    @value.setter
    def value(self, value):
        if self.template and value and not self.template.bounds.validate(value):
            raise ValueError("the value is inconsistent with the template")

        if value is None:
            self._value = None
        elif isinstance(value, (BaseValue, str, bool)):
            self._value = value
        else:
            raise ValueError("value must be a BaseValue, string or bool")

    @property
    def template(self):
        """Get template."""
        return self._template

    @template.setter
    def template(self, template):
        if template is None:
            self._template = None
        elif isinstance(template, LinkByUID):
            self._template = template
        elif isinstance(template, AttributeTemplate):
            if not template.validate(self):
                raise ValueError("Template is incompatible with attr {}".format(self.name))
            if self.value and not template.bounds.validate(self.value):
                raise ValueError("the template is inconsistent with the value")
            self._template = template
        else:
            raise ValueError("template must be a BaseAttributeTemplate or LinkByUID")

    @property
    def origin(self):
        """Get origin."""
        return self._origin

    @origin.setter
    def origin(self, origin):
        if origin is None:
            raise ValueError("origin must be specified (but may be `unknown`)")
        self._origin = Origin.get_value(origin)

    @property
    def file_links(self):
        """Get file links."""
        return self._file_links

    @file_links.setter
    def file_links(self, file_links):
        self._file_links = validate_list(file_links, FileLink)



"""Base class for all bounds."""
from abc import abstractmethod

from taurus.entity.dict_serializable import DictSerializable
from taurus.entity.value.base_value import BaseValue


class BaseBounds(DictSerializable):
    """Base class for bounds, including RealBounds and CategoricalBounds."""

    @abstractmethod
    def validate(self, value):
        """
        Check if a value is a member of this bounds.

        Parameters
        ----------
        value: BaseValue
            Value to validate

        Returns
        -------
        bool
            True if the value is a member of the bounds, and False otherwise

        """
        if value is None:
            return False
        if isinstance(value, BaseValue):
            return True
        raise TypeError('{} is not a Value object'.format(value))

    @abstractmethod
    def contains(self, bounds):
        """
        Check if another bounds is contained within this bounds.

        Parameters
        ----------
        bounds: BaseBounds
            Other bounds object to check.

        Returns
        -------
        bool
            True if any value that validates true for bounds also validates true for this

        """
        if bounds is None:
            return False
        if isinstance(bounds, BaseBounds):
            return True
        raise TypeError('{} is not a Bounds object'.format(bounds))



"""Base class for all entities."""
from taurus.entity.dict_serializable import DictSerializable
from taurus.entity.case_insensitive_dict import CaseInsensitiveDict


class BaseEntity(DictSerializable):
    """
    Base class for any entity, which includes objects and templates.

    Parameters
    ----------
    uids: Map[str, str]
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str]
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.

    """

    typ = "base"

    def __init__(self, uids, tags):
        self._tags = None
        self.tags = tags

        self._uids = None
        self.uids = uids

    def content_hash(self):
        """A hash of the object's content."""
        return str(sorted(list(self.__dict__.items())))

    @property
    def tags(self):
        """Get the tags."""
        return self._tags

    @tags.setter
    def tags(self, tags):
        if tags is None:
            self._tags = []
        elif isinstance(tags, list):
            self._tags = tags
        else:
            self._tags = [tags]

    @property
    def uids(self):
        """Get the uids."""
        return self._uids

    @uids.setter
    def uids(self, uids):
        if uids is None:
            self._uids = CaseInsensitiveDict()
        elif isinstance(uids, dict):
            self._uids = CaseInsensitiveDict(**uids)
        else:
            self._uids = CaseInsensitiveDict(**{uids[0]: uids[1]})

    def add_uid(self, scope, uid):
        """
        Add a uid.

        Parameters
        ----------
        scope: str
            scope of the uid
        uid: str
            Unique identifier

        """
        self.uids[scope] = uid



from taurus.entity.attribute.base_attribute import BaseAttribute
from taurus.enumeration import AttributeType


class Parameter(BaseAttribute):
    """
    Parameter of a process or measurement.

    Parameters are the non-environmental variables (typically specified and controlled) that may
    affect a process or measurement: e.g. oven dial temperature for a kiln firing, magnification
    for a measurement taken with an electron microscope.

    """

    typ = "parameter"
    attribute_type = AttributeType.PARAMETER



"""A nominal composition value."""
from taurus.entity.value.composition_value import CompositionValue


class NominalComposition(CompositionValue):
    """
    Nominal composition, represented as a map from the component names to the quantities.

    The quantities do not express an uncertainty but also do not imply that there is absolute
    certainty to their values.

    Parameters
    ----------
    quantities: Map[String, Number]
        A map from each component to its amount.  The quantities are not required to be expressed
        on a unit or fractional basis. The following are all acceptable:

        * dict(acetone=0.25, methanol=0.75)
        * dict(acetone=1, methanol=3)
        * dict(acetone=3.5, methanol=10.5)

    """

    typ = "nominal_composition"

    def __init__(self, quantities=None):
        self._quantities = None
        self.quantities = quantities

    @property
    def quantities(self):
        """Get a map from the components to their quantities."""
        return self._quantities

    @quantities.setter
    def quantities(self, quantities):
        if quantities is None:
            self._quantities = {}
        elif isinstance(quantities, dict):
            self._quantities = quantities
        elif isinstance(quantities, list):
            self._quantities = dict(quantities)
        else:
            raise ValueError("quantities must be dict or None")

    def as_dict(self):
        """
        Convert the composition to a dictionary.

        Overrides the ordinary
        :func:`as_dict() <taurus.entity.dict_serializable.DictSerializable.as_dict>` method in
        that the `quantities` field is turned into a list of lists,
        each of which has the form [component, quantity].
        """
        return {"type": self.typ, "quantities": list(list(x) for x in self.quantities.items())}

    @property
    def components(self):
        """Get a list of the components."""
        return self._quantities.keys()



"""A value that nominally is equal to a single category."""
from taurus.entity.setters import validate_str
from taurus.entity.value.categorical_value import CategoricalValue


class NominalCategorical(CategoricalValue):
    """
    A nominal category that the value is believe to have. It may not be exact.

    Parameters
    ----------
    category: str
        The nominal category.

    """

    typ = "nominal_categorical"

    def __init__(self, category=None):
        self._category = category

    @property
    def category(self):
        """Get the category."""
        return self._category

    @category.setter
    def category(self, category):
        if category is None:
            self._category = None
        self._category = validate_str(category)



"""An empirical chemical formaula."""
from typing import Dict

from taurus.entity.value.composition_value import CompositionValue
from pymatgen.core.composition import Composition as ElementalComposition
from pymatgen.core.composition import Element


class EmpiricalFormula(CompositionValue):
    """
    An empirical chemical formula where only the relative stoichiometries matter.

    Parameters
    ----------
    formula: str
        A string corresponding to the chemical formula of the composition.
        It must be parseable by pymatgen. The order and grouping of the elements is ignored.

    """

    typ = "empirical_formula"
    skip = {"_quantities"}

    def __init__(self, formula=None):
        self._formula = None
        self._quantities = None
        self.formula = formula

    @property
    def formula(self):
        """Get the formula as a string."""
        return self._formula

    @formula.setter
    def formula(self, value):
        self._quantities = self._parse_formula(value)
        self._formula = value

    @staticmethod
    def _parse_formula(formula_str) -> Dict[str, float]:
        """Parse the formula (a string) to return a map from elements to amounts."""
        return ElementalComposition(formula_str).to_reduced_dict

    @property
    def components(self):
        """Get all components of the chemical formula."""
        return self._quantities.keys()

    @property
    def quantities(self):
        """Get a map from components to quantities."""
        return self._quantities

    @staticmethod
    def all_elements():
        """The set of all elements in the periodic table."""
        return {e.value for e in Element}



from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_conditions import HasConditions
from taurus.entity.object.has_properties import HasProperties
from taurus.entity.object.has_parameters import HasParameters
from taurus.entity.object.has_source import HasSource
from taurus.entity.setters import validate_list
from taurus.entity.valid_list import ValidList


class MeasurementRun(BaseObject, HasConditions, HasProperties, HasParameters, HasSource):
    """
    A measurement run.

    This contains a link to the material the measurement is performed on, as well as links to
    any properties, conditions, and parameters.

    Parameters
    ----------
    name: str, optional
        Name of the measurement run.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the measurement run.
    conditions: List[Condition], optional
        Conditions under which this measurement run occurs.
    parameters: List[Parameter], optional
        Parameters of this measurement run.
    properties: List[Property], optional
        Properties that are measured during this measurement run.
    spec: MeasurementSpec
        The measurement specification of which this is an instance.
    material: MaterialRun
        The material run being measured.
    spec: MaterialSpec
        The material specification of which this is an instance.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.
    source: PerformedSource, optional
        Information about the person who performed the run and when.

    """

    typ = "measurement_run"

    def __init__(self, name=None, spec=None, material=None,
                 properties=None, conditions=None, parameters=None,
                 uids=None, tags=None, notes=None, file_links=None, source=None):
        BaseObject.__init__(self, name=name, uids=uids, tags=tags, notes=notes,
                            file_links=file_links)
        HasProperties.__init__(self, properties)
        HasConditions.__init__(self, conditions)
        HasParameters.__init__(self, parameters)
        HasSource.__init__(self, source)

        self._material = None
        self.material = material
        self._spec = None
        self.spec = spec

    @property
    def material(self):
        """Get the material."""
        return self._material

    @material.setter
    def material(self, value):
        from taurus.entity.object import MaterialRun
        from taurus.entity.link_by_uid import LinkByUID
        if value is None:
            self._material = value
        elif isinstance(value, MaterialRun):
            self._material = value
            if not isinstance(value.measurements, ValidList):
                value._measurements = validate_list(self, MeasurementRun)
            else:
                value._measurements.append(self)
        elif isinstance(value, LinkByUID):
            self._material = value
        else:
            raise ValueError("material must be a MaterialRun or LinkByUID: {}".format(value))

    @property
    def spec(self):
        """Get the measurement spec."""
        return self._spec

    @spec.setter
    def spec(self, spec):
        from taurus.entity.object.measurement_spec import MeasurementSpec
        from taurus.entity.link_by_uid import LinkByUID
        if spec is None:
            self._spec = None
        elif isinstance(spec, (MeasurementSpec, LinkByUID)):
            self._spec = spec
        else:
            raise ValueError("spec must be a MeasurementSpec: {}".format(spec))

    @property
    def template(self):
        """Get the template of the spec, if applicable."""
        from taurus.entity.object.measurement_spec import MeasurementSpec
        if isinstance(self.spec, MeasurementSpec):
            return self.spec.template
        else:
            return None



from abc import ABC
import json


class DictSerializable(ABC):
    """A base class for objects that can be represented as a dictionary and serialized."""

    typ = NotImplemented
    skip = set()

    @classmethod
    def from_dict(cls, d):
        """
        Reconstitute the object from a dictionary.

        Parameters
        ----------
        d: dict
            The object as a dictionary of key-value pairs that correspond to the object's fields.

        Returns
        -------
        DictSerializable
            The deserialized object.

        """
        # noinspection PyArgumentList
        # DictSerializable's constructor is not intended for use,
        # but all of its children will use from_dict like this.
        return cls(**d)

    def as_dict(self):
        """
        Convert the object to a dictionary.

        Returns
        -------
        dict
            A dictionary representation of the object, where the keys are its fields.

        """
        keys = {x.lstrip('_') for x in vars(self) if x not in self.skip}
        attributes = {k: self.__getattribute__(k) for k in keys}
        attributes["type"] = self.typ
        return attributes

    def dump(self):
        """
        Convert the object to a JSON dictionary, so that every entry is serialized.

        Uses the json encoder client, so objects with uids are converted to LinkByUID dictionaries.

        Returns
        -------
        str
            A string representation of the object as a dictionary.

        """
        from taurus.client.json_encoder import dumps
        return json.loads(dumps(self))[1]

    @staticmethod
    def build(d):
        """
        Build an object from a JSON dictionary.

        This differs from `from_dict` in that the values themselves may *also* be dictionaries
        corresponding to serialized DictSerializable objects.

        Parameters
        ----------
        d: dict
            The object as a serialized dictionary.

        Returns
        -------
        DictSerializable
            The deserialized object.

        """
        from taurus.client.json_encoder import loads, dumps
        return loads(dumps(d))

    def __repr__(self):
        return str(self.as_dict())

    def __eq__(self, other):
        if isinstance(other, DictSerializable):
            self_dict = self.as_dict()
            other_dict = other.as_dict()
            return self_dict == other_dict
        else:
            return False

    # TODO make a hash function which reflects __eq__?
    def __hash__(self):
        return super().__hash__()



"""An example ingest of a material run."""
from taurus import units
from taurus.entity.attribute.condition import Condition
from taurus.entity.attribute.parameter import Parameter
from taurus.entity.attribute.property import Property
from taurus.entity.bounds.categorical_bounds import CategoricalBounds
from taurus.entity.bounds.real_bounds import RealBounds
from taurus.entity.object import MaterialRun, MeasurementRun
from taurus.entity.template.condition_template import ConditionTemplate
from taurus.entity.template.parameter_template import ParameterTemplate
from taurus.entity.template.property_template import PropertyTemplate
from taurus.entity.value.discrete_categorical import DiscreteCategorical
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.value.normal_real import NormalReal

known_properties = {
    "density": PropertyTemplate(
        name="density",
        bounds=RealBounds(lower_bound=0.0, upper_bound=1000.0, default_units='g / cm^3')
    ),
    "kinematic viscosity": PropertyTemplate(
        name="kinematic viscosity",
        bounds=RealBounds(lower_bound=0.0, upper_bound=10.0**40, default_units="m^2 / s")
    )
}

known_conditions = {
    "temperature": ConditionTemplate(
        name="temperature",
        bounds=RealBounds(lower_bound=0.0, upper_bound=1000.0, default_units='K')
    )
}

known_parameters = {
    "knob_2_setting": ParameterTemplate(
        name="knob_2_setting",
        bounds=CategoricalBounds(categories={"low", "medium", "high"})
    )
}


def _parse_value(val):
    """Example field-parsing logic."""
    # If the string is complicated, split it up and try to get uncertainty and/or units
    if isinstance(val, str) and len(val.split()) > 1:
        toks = val.split()
        mean = float(toks[0])
        std = -1
        if toks[1] in {"+-", "+/-"}:
            std = float(toks[2])

        try:
            unit = units.parse_units(toks[-1])
        except (ValueError, units.UndefinedUnitError):
            print("Couldn't find {}".format(toks[-1]))
            unit = ''

        if std >= 0:
            return NormalReal(mean=mean, std=std, units=unit)
        else:
            return NominalReal(mean, units=unit)
    # if it is just a number wrap it in a nominal value
    elif isinstance(val, (float, int)):
        return NominalReal(val, '')
    # if it is a single string, its either a single number of a category
    elif isinstance(val, str):
        try:
            num = float(val)
            return NominalReal(num, '')
        except ValueError:
            return DiscreteCategorical(val)
    else:
        raise ValueError("Couldn't parse {}".format(val))


def ingest_material_run(data, material_spec=None, process_run=None):
    """Ingest material run with data, a material spec, and an originating process run."""
    if isinstance(data, list):
        return [ingest_material_run(x, material_spec) for x in data]

    if not isinstance(data, dict):
        raise ValueError("This ingester operates on dict, but got {}".format(type(data)))

    material = MaterialRun()

    sample_id = data.get("sample_id")
    if sample_id:
        material.add_uid("given_sample_id", sample_id)

    tags = data.get("tags")
    if tags:
        material.tags = tags

    for experiment in data.get("experiments", []):
        measurement = MeasurementRun()

        for name in set(known_properties.keys()).intersection(experiment.keys()):
            prop = Property(
                name=name,
                template=known_properties[name],
                value=_parse_value(experiment[name])
            )
            measurement.properties.append(prop)

        for name in set(known_conditions.keys()).intersection(experiment.keys()):
            cond = Condition(
                name=name,
                template=known_conditions[name],
                value=_parse_value(experiment[name])
            )
            measurement.conditions.append(cond)

        for name in set(known_parameters.keys()).intersection(experiment.keys()):
            param = Parameter(
                name=name,
                template=known_parameters[name],
                value=_parse_value(experiment[name])
            )
            measurement.parameters.append(param)

        scan_id = experiment.get("scan_id")
        if scan_id:
            measurement.add_uid("given_scan_id", scan_id)

        tags = experiment.get("tags")
        if tags:
            measurement.tags = tags

        measurement.material = material

    if material_spec:
        material.material_spec = material_spec

    if process_run:
        material.process = process_run

    return material



"""A nominal real value."""
from taurus.entity.value.continuous_value import ContinuousValue


class NominalReal(ContinuousValue):
    """
    Nominal real, which does not specify an uncertainty but is not to be assumed exact.

    Parameters
    ----------
    nominal: float
        A nominal value--not assumed to be exact.
    units: str
        A string describing the units. Units must be present and they must be parseable by Pint.
        An empty string can be used for the units of a dimensionless quantity.

    """

    typ = "nominal_real"

    def __init__(self, nominal=None, units=None):
        ContinuousValue.__init__(self, units)
        assert isinstance(nominal, (int, float)), \
            "nominal value must be an int or float"
        self.nominal = float(nominal)



"""For entities that have parameters."""
from taurus.entity.source.performed_source import PerformedSource


class HasSource(object):
    """Mixin-trait for entities that include sources (data provenance)."""

    def __init__(self, source):
        self._source = None
        self.source = source

    @property
    def source(self):
        """Get the list of parameters."""
        return self._source

    @source.setter
    def source(self, value):
        if value is None:
            self._source = None
        elif isinstance(value, PerformedSource):
            self._source = value
        else:
            raise TypeError("Source must be a PerformedSource; was {}".format(type(value)))



from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_parameters import HasParameters
from taurus.entity.object.has_conditions import HasConditions
from taurus.entity.object.has_template import HasTemplate


class MeasurementSpec(BaseObject, HasParameters, HasConditions, HasTemplate):
    """
    A measurement specification.

    This includes links to the conditions and parameters under which the measurement is
    expected to be performed.

    Parameters
    ----------
    name: str, optional
        Name of the measurement spec.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the measurement spec.
    conditions: List[Condition], optional
        Conditions under which this measurement spec occurs.
    parameters: List[Parameter], optional
        Parameters of this measurement spec.
    template: MeasurementTemplate
        A template bounding the valid values for the measurement's properties, parameters,
        and conditions.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    """

    typ = "measurement_spec"

    def __init__(self, name=None, template=None,
                 parameters=None, conditions=None,
                 uids=None, tags=None, notes=None, file_links=None):
        BaseObject.__init__(self, name=name, uids=uids, tags=tags, notes=notes,
                            file_links=file_links)
        HasParameters.__init__(self, parameters=parameters)
        HasConditions.__init__(self, conditions=conditions)
        HasTemplate.__init__(self, template=template)



"""Discrete distribution across several categories."""
from toolz import keymap

from taurus.entity.setters import validate_str
from taurus.entity.value.categorical_value import CategoricalValue


class DiscreteCategorical(CategoricalValue):
    """
    Distribution over a discrete set of categories.

    Parameters
    ----------
    probabilities: str or Map[str, float]
        The categories and their probabilities.

        If a string is provided, that string corresponds to the only category and is given a
        probability of 1.0

        If a dictionary is provided, then each key is a category and its value is the probability
        of that category. The probabilities *must* sum to 1.0.

    """

    typ = "discrete_categorical"

    def __init__(self, probabilities=None):
        self._probabilities = None
        self.probabilities = probabilities

    @property
    def probabilities(self):
        """Get the map from categories to probabilities."""
        return self._probabilities

    @probabilities.setter
    def probabilities(self, probabilities):
        if probabilities is None:
            self._probabilities = None
        elif isinstance(probabilities, str):
            self._probabilities = {validate_str(probabilities): 1.0}
        elif isinstance(probabilities, dict):
            if abs(sum(probabilities.values()) - 1.0) > 1.0e-9:
                raise ValueError("probabilities must sum to 1.0")

            self._probabilities = keymap(validate_str, probabilities)
        else:
            raise ValueError("probabilities must be dict or single value")



"""For entities that have a property template."""
from taurus.entity.setters import validate_list
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.template.property_template import PropertyTemplate


class HasPropertyTemplates(object):
    """
    Mixin-trait for entities that include property templates.

    Parameters
    ----------
    properties: List[PropertyTemplate]
        A list of this entity's property templates.

    """

    def __init__(self, properties):
        self._properties = None
        self.properties = properties

    @property
    def properties(self):
        """
        Get the list of property templates.

        Returns
        -------
        List[PropertyTemplate]
            List of this entity's property templates

        """
        return self._properties

    @properties.setter
    def properties(self, properties):
        # make sure the properties are a list
        lst = validate_list(properties, (PropertyTemplate, list, tuple))

        # make sure attribute can be a Property
        # TODO: list.map(_.validate_scope(AttributeType.PROPERTY)) all true

        # convert any templates into (template, bounds) pairs and
        # validate that any (template, bounds) pairs are consistent
        self._properties = list(map(BaseTemplate._homogenize_ranges, lst))

    def validate_properties(self, obj):
        """Validate an object's properties against its templates."""
        # validate the template against the object's properties
        BaseTemplate._validate_attributes(self, obj, "properties")



"""Attribute templates."""
from taurus.entity.base_entity import BaseEntity
from taurus.entity.bounds.base_bounds import BaseBounds


class AttributeTemplate(BaseEntity):
    """
    An attribute template, which can be a property, parameter, or condition template.

    Parameters
    ----------
    name: str
        The name of the attribute template.
    bounds: :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`
        Bounds circumscribe the values that are valid according to this attribute template.
    description: str, optional
        A long-form description of the attribute template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.

    """

    def __init__(self, name=None, description=None, bounds=None, uids=None, tags=None):
        BaseEntity.__init__(self, uids, tags)
        self.name = name
        self.description = description

        if not isinstance(name, str):
            raise ValueError("name must be a string")

        self._bounds = None
        self.bounds = bounds

    @property
    def bounds(self):
        """Get the bounds."""
        return self._bounds

    @bounds.setter
    def bounds(self, value):
        if value is None:
            raise ValueError("Bounds are required on AttributeTemplate")
        if not isinstance(value, BaseBounds):
            raise ValueError("Bounds must be an instance of BaseBounds")
        self._bounds = value

    def validate(self, attribute):
        """
        Check that the attribute is consistent with this template.

        This method returns True/False instead of throwing a ValueError.  If the name or value
        of the attribute is not defined, it is not checked (i.e. default True).

        :param attribute: to check the validity of
        :return: True if valid and False if invalid
        """
        if attribute.name != self.name:
            return False

        if attribute.value is not None and not self.bounds.validate(attribute.value):
            return False

        return True



# flake8: noqa
from .material_run import MaterialRun
from .measurement_run import MeasurementRun
from .process_run import ProcessRun
from .material_spec import MaterialSpec
from .measurement_spec import MeasurementSpec
from .process_spec import ProcessSpec



"""A list that can validate its contents."""
from collections.abc import Iterable


class ValidList(list):
    """
    A list-like class that verifies that its content conforms to specified types.

    Parameters
    ----------
    _list: Iterable
        The initial values for the elements of the list.
    content_type: Type or Iterable[Type]
        The allowed type(s) for the content of the list.
    trigger: function
        A function that gets invoked whenever a new element is added.

    """

    _content_type = tuple([])

    def __init__(self, _list, content_type=None, trigger=None):
        if content_type is None:
            content_type = tuple()

        if isinstance(content_type, dict):
            raise TypeError('A dict is not an acceptable container for content filters')
        elif isinstance(content_type, Iterable):
            self._content_type = tuple(content_type)
        else:
            self._content_type = tuple([content_type])
        for elem in self._content_type:
            if not isinstance(elem, type):
                raise TypeError('Content filters must be types')
        for value in _list:
            self._validate(value)
        self._trigger = None
        if trigger is not None:
            if not callable(trigger):
                raise TypeError('Triggers must be callable')
            self._trigger = trigger
            for value in _list:
                self._trigger(self, value)
        list.__init__(self, _list)

    def _validate(self, value):
        """
        Validate a value against the allowed types.

        Parameters
        ----------
        value: Any
            The value to validate.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If `value` is not one of the allowed types.

        """
        if not isinstance(value, self._content_type):
            raise TypeError(
                'Value is not of an accepted type: {} =/= {}'.format(value, self._content_type))

    def __setitem__(self, index, value):
        """
        Called to implement assignment to self[index].

        Validates that `value` is one of the allowed types.

        Parameters
        ----------
        index: int
            The index of the element of the list to change
        value: Any
            The value to set the element to

        Returns
        -------
        None
            `value` is inserted into the list at position `index`, if it is valid.

        """
        self._validate(value)
        if self._trigger is not None:
            self._trigger(self, value)
        super().__setitem__(index, value)

    def append(self, value):
        """
        Add an item to the end of the list; equivalent to a[len(a):] = [x].

        Validates that `value` is one of the allowed types.

        Parameters
        ----------
        value: Any
            The value to append at the end of the list.

        Returns
        -------
        None
            `value` is appended at the end of the list, if it is valid.

        """
        self._validate(value)
        if self._trigger is not None:
            self._trigger(self, value)
        super().append(value)

    def extend(self, list_):
        """
        Extend the list by appending all the items in the given list; equivalent to a[len(a):] = L.

        Validates that `value` is one of the allowed types.

        Parameters
        ----------
        list_: list
            The list of values to append at the end of the list.

        Returns
        -------
        None
            `list_` is appended at the end of the list, if all its entries are valid.

        """
        if isinstance(list_, Iterable):
            for value in list_:
                self._validate(value)
        else:
            raise TypeError("'{}' object is not iterable".format(type(list_)))
        if self._trigger is not None:
            for value in list_:
                self._trigger(self, value)
        super().extend(list_)

    def insert(self, i, value):
        """
        Insert a value at a given position, if it is one of the allowed types.

        a.insert(0, x) inserts at the front of the list, and a.insert(len(a), x)
        is equivalent to a.append(x).

        Parameters
        ----------
        i: int
            The index of the element before which to insert.
        value: Any
            The value to insert into the list.

        Returns
        -------
        None
            `value` is inserted into the list at position `i`, if it is valid.

        """
        self._validate(value)
        if self._trigger is not None:
            self._trigger(self, value)
        super().insert(i, value)



"""For entities that have parameters."""
from taurus.entity.attribute.parameter import Parameter
from taurus.entity.setters import validate_list


class HasParameters(object):
    """Mixin-trait for entities that include parameters."""

    def __init__(self, parameters):
        self._parameters = None
        self.parameters = parameters

    @property
    def parameters(self):
        """Get the list of parameters."""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = validate_list(parameters, Parameter)



"""Base template."""
from taurus.entity.base_entity import BaseEntity
from taurus.entity.bounds.base_bounds import BaseBounds
from taurus.entity.link_by_uid import LinkByUID
from taurus.entity.template.attribute_template import AttributeTemplate


class BaseTemplate(BaseEntity):
    """
    Base class for all object templates.

    Parameters
    ----------
    name: str, optional
        The name of the object template.
    description: str, optional
        Long-form description of the object template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.

    """

    def __init__(self, name=None, description=None, uids=None, tags=None):
        BaseEntity.__init__(self, uids, tags)
        self.name = name
        self.description = description

    def validate(self, obj):
        """Validate an object against the template."""
        raise NotImplementedError(
            "Subclass of BaseTemplate didn't implement validate: {}".format(type(self)))

    @staticmethod
    def _homogenize_ranges(template_or_tuple):
        """
        Take either a template or pair and turn it into a (template, bounds) pair.

        If no bounds are provided, use the attribute template's default bounds.

        Parameters
        ----------
        template_or_tuple: AttributeTemplate OR a list or
        tuple [AttributeTemplate or LinkByUID, BaseBounds]
           An attribute template, optionally with another Bounds object that is more
           restrictive than the attribute template's default bounds.

        Returns
        -------
        List[AttributeTemplate or LinkByUID, BaseBounds]
            The attribute template and bounds that should be applied the the attribute
            when used in the context of **this** object.

        """
        # if given a template, pull out its bounds
        if isinstance(template_or_tuple, AttributeTemplate):
            return [template_or_tuple, template_or_tuple.bounds]
        # if given a (template, bounds) pair,
        # check that the bounds is consistent with that of the template
        elif isinstance(template_or_tuple, (tuple, list)):
            first, second = template_or_tuple
            if isinstance(first, LinkByUID) and isinstance(second, BaseBounds):
                return [first, second]
            if isinstance(first, AttributeTemplate) and isinstance(second, BaseBounds):
                if first.bounds.contains(second):
                    return [first, second]
                else:
                    raise ValueError("Range and template are inconsistent")
        raise ValueError("Expected a template or (template, bounds) tuple")

    @staticmethod
    def _validate_attributes(template, obj, attr_name):
        """Validate a specific attribute in an object against its template."""
        # Skip validation if the object doesn't contain the attribute,
        # e.g. MaterialRun doesn't contain properties
        if not hasattr(obj, attr_name):
            return

        # check each contained (template, bounds)
        for (attr_template, bounds) in getattr(template, attr_name):

            # find any attributes that match the name
            name = getattr(attr_template, 'name', None)
            for attribute in filter(lambda x: x.name == name, getattr(obj, attr_name)):
                # make sure the value is within the accepted range (the bounds)
                if not bounds.validate(attribute.value):
                    raise RuntimeWarning(
                        "Template and object do not match: attribute '{attr_name}' has value "
                        "{val}, which is invalid against bounds {bounds}".format(
                            attr_name=attribute.name, val=attribute.value, bounds=bounds)
                    )

                # if the attribute doesn't have a template, assign this template to it
                if attribute.template is None:
                    attribute.template = attr_template
                # if it does have a template and its not a LinkByUID, make sure its the right one
                elif attribute.template != attr_template \
                        and isinstance(attribute.template, AttributeTemplate) \
                        and isinstance(attr_template, AttributeTemplate):
                    raise ValueError(
                        "Template and object templates don't match for all {}".format(attr_name))



from taurus.entity.attribute.base_attribute import BaseAttribute
from taurus.enumeration import AttributeType


class Property(BaseAttribute):
    """
    Property of a material, measured in a MeasurementRun or specified in a MaterialSpec.

    Properties are characteristics of a material that could be measured, e.g. chemical composition,
     density, yield strength.

    """

    typ = "property"
    attribute_type = AttributeType.PROPERTY



from taurus.entity.attribute.property_and_conditions import PropertyAndConditions
from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_template import HasTemplate
from taurus.entity.setters import validate_list


class MaterialSpec(BaseObject, HasTemplate):
    """
    A material specification.

    This includes a link to the originating process and specified properties with conditions.

    Parameters
    ----------
    name: str, optional
        Name of the material spec.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the material spec.
    process: ProcessSpec
        Process that produces this material.
    properties: List[PropertyAndConditions], optional
        Properties of the material, along with an optional list of conditions under which
        those properties are measured.
    template: MaterialTemplate, optional
        A template bounding the valid values for this material's properties.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    """

    typ = "material_spec"

    def __init__(self, name=None, template=None,
                 properties=None, process=None, uids=None, tags=None,
                 notes=None, file_links=None):
        BaseObject.__init__(self, name=name, uids=uids, tags=tags, notes=notes,
                            file_links=file_links)
        self._properties = None
        self.properties = properties
        self._process = None
        self.process = process
        HasTemplate.__init__(self, template)

    @property
    def properties(self):
        """Get the list of property-and-conditions."""
        return self._properties

    @properties.setter
    def properties(self, properties):
        self._properties = validate_list(properties, PropertyAndConditions)

    @property
    def process(self):
        """Get the originating process spec."""
        return self._process

    @process.setter
    def process(self, process):
        """
        Link to the ProcessSpec that creates this MaterialSpec.

        If the input, process, is not an instance of ProcessSpec, raise an error.
        Otherwise, make a bi-directional link: this MaterialSpec is linked to
        process, and process has its output_material field linked to this MaterialSpec
        """
        from taurus.entity.object.process_spec import ProcessSpec
        from taurus.entity.link_by_uid import LinkByUID
        if process is None:
            self._process = None
        elif isinstance(process, LinkByUID):
            self._process = process
        elif isinstance(process, ProcessSpec):
            process._output_material = self
            self._process = process
        else:
            raise ValueError("process must be an instance of ProcessSpec; "
                             "instead received type {}: {}".format(type(process), process))



"""Represents a link to an external file."""
from taurus.entity.dict_serializable import DictSerializable


class FileLink(DictSerializable):
    """
    FileLink stores a name and link to an external resource.

    More information can be found in the
    `data model documentation \
    <https://citrineinformatics.github.io/taurus-documentation/specification/file-links/>`_

    Once the file protocol is defined, there should be substantial validation.

    Parameters
    ----------
    filename: str
        The name of the file.
    url: str
        URL that can be used to access the file.

    """

    typ = "file_link"

    def __init__(self, filename, url):
        DictSerializable.__init__(self)
        self.filename = filename
        self.url = url



"""Demonstrate attaching measurements to a material."""
import random
import string

from taurus.entity.attribute.property import Property
from taurus.entity.object import MeasurementRun
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.value.normal_real import NormalReal
from taurus.enumeration import Origin

# recommended values taken from
# https://www.shimadzu.com/an/industry/petrochemicalchemical/n9j25k00000pyv3w.html
thickness: float = 4.0  # mm
length = 80.0  # mm
width = 10.0  # mm
span = 64.0  # mm
punch_radius = 5.0  # mm
support_radius = 5.0  # mm
applied_force = 100.0  # N


def __random_my_id():
    """Create random 8-letter id."""
    return "".join([random.choice(string.ascii_lowercase) for _ in range(8)])


def make_demo_measurements(num_measurements, extra_tags=frozenset()):
    """Make a measurement object."""
    return [
        make_flexural_test_measurement(
            my_id=__random_my_id(),
            deflection=random.random(),
            extra_tags=extra_tags
        ) for _ in range(num_measurements)
    ]


def make_flexural_test_measurement(my_id, deflection, extra_tags=frozenset()):
    """
    Compute the stree, strain, and modulus.

    According to https://en.wikipedia.org/wiki/Three-point_flexural_test
    """
    stress = 3 * applied_force * span / (2 * thickness * thickness * width)
    strain = 6 * deflection * thickness / (span * span)
    modulus = stress / strain

    measurement = MeasurementRun(
        uids={"my_id": my_id},
        tags=["3_pt_bend", "mechanical", "flex"] + list(extra_tags),
        properties=[
            Property(
                name="flexural stress",
                value=NormalReal(stress, std=(0.01 * stress), units="MPa"),
                origin=Origin.MEASURED
            ),
            Property(
                name="flexural strain",
                value=NormalReal(strain, std=(0.01 * strain), units=""),
                origin=Origin.MEASURED
            ),
            Property(
                name="flexural modulus",
                value=NormalReal(modulus, std=(0.01 * modulus), units="MPa"),
                origin=Origin.MEASURED
            ),
            Property(
                name="deflection",
                value=NominalReal(deflection, units="mm"),
                origin=Origin.MEASURED
            )
        ]
    )
    return measurement



"""
Metadata class.

THIS IS CURRENTLY UNUSED.
We might use it in the future, but its not currently part of the spec
"""

from taurus.entity.attribute.base_attribute import BaseAttribute
from taurus.enumeration import AttributeType


class Metadata(BaseAttribute):
    """Ancillary information about a material or process."""

    typ = "metadata"
    attribute_type = AttributeType.METADATA



"""Implementation of units."""
import pint
import pkg_resources
from pint import UnitRegistry
from pint.quantity import _Quantity
from pint.unit import _Unit

try:
    str
except NameError:
    str = str

# use the default unit registry for now
_ureg = UnitRegistry(filename=pkg_resources.resource_filename("taurus.units", "citrine_en.txt"))


# alias the error that is thrown when units are incompatible
# this helps to isolate the dependence on pint
IncompatibleUnitsError = pint.errors.DimensionalityError
UndefinedUnitError = pint.errors.UndefinedUnitError


def _unit_to_str(unit):
    """Helper that pulls a string representation of the unit from a quantity."""
    if not isinstance(unit, _Quantity):
        raise ValueError("Expecting a unit or quantity")
    return str(unit.units)


def parse_units(units):
    """Parse a string or _Unit into a standard string representation of the unit."""
    if units is None:
        return None
    elif units == '':
        return 'dimensionless'
    elif isinstance(units, str):
        return _unit_to_str(_ureg(units))
    elif isinstance(units, _Unit):
        return _unit_to_str(units)
    else:
        raise UndefinedUnitError("Units must be given as a recognized unit string or Units object")


def convert_units(value, starting_unit, final_unit):
    """
    Convert the value from the starting_unit to the final_unit.

    :param value: magnitude to convert (number)
    :param starting_unit: unit that the magnitude is currently in (str)
    :param final_unit: unit that the magnitude should be returned in (str)
    """
    return _ureg.Quantity(value, starting_unit).to(final_unit).magnitude



"""For entities that have templates."""
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.link_by_uid import LinkByUID


class HasTemplate(object):
    """Mix-in trait for objects that can be assigned templates."""

    def __init__(self, template=None):
        self._template = None
        self.template = template

    @property
    def template(self):
        """Get the template."""
        return self._template

    @template.setter
    def template(self, template):
        if template is None:
            self._template = None
        elif isinstance(template, BaseTemplate):
            # run the template validation when the template is assigned
            template.validate(self)
            self._template = template
        elif isinstance(template, LinkByUID):
            # LinkByUID cannot be validated, but is still assigned
            self._template = template
        else:
            raise ValueError("Template must be a template")



"""For entities that hve quantities."""
from taurus.entity.bounds.real_bounds import RealBounds
from taurus.entity.value.continuous_value import ContinuousValue
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.value.normal_real import NormalReal
from taurus.entity.value.uniform_real import UniformReal

fraction_bounds = RealBounds(lower_bound=0.0, upper_bound=1.0, default_units='')


class HasQuantities(object):
    """Mixin-trait that includes the mass, volume, number fraction, and absolute quantity."""

    def __init__(self, mass_fraction=None, volume_fraction=None, number_fraction=None,
                 absolute_quantity=None):

        self._mass_fraction = None
        self.mass_fraction = mass_fraction

        self._volume_fraction = None
        self.volume_fraction = volume_fraction

        self._number_fraction = None
        self.number_fraction = number_fraction

        self._absolute_quantity = None
        self.absolute_quantity = absolute_quantity

    @property
    def mass_fraction(self):
        """Get mass fraction."""
        return self._mass_fraction

    @mass_fraction.setter
    def mass_fraction(self, mass_fraction):
        if mass_fraction is not None and not fraction_bounds.validate(mass_fraction):
            raise ValueError("mass_fraction was not given as a fraction: {}".format(mass_fraction))
        self._mass_fraction = mass_fraction

    @property
    def volume_fraction(self):
        """Get volume fraction."""
        return self._volume_fraction

    @volume_fraction.setter
    def volume_fraction(self, volume_fraction):
        if volume_fraction is not None and not fraction_bounds.validate(volume_fraction):
            raise ValueError("volume_fraction was not given as a fraction")
        self._volume_fraction = volume_fraction

    @property
    def number_fraction(self):
        """Get number fraction."""
        return self._number_fraction

    @number_fraction.setter
    def number_fraction(self, number_fraction):
        if number_fraction is not None and not fraction_bounds.validate(number_fraction):
            raise ValueError("number_fraction was not given as a fraction")
        self._number_fraction = number_fraction

    @property
    def absolute_quantity(self):
        """Get absolute quantity."""
        return self._absolute_quantity

    @absolute_quantity.setter
    def absolute_quantity(self, absolute_quantity):
        if absolute_quantity is None:
            self._absolute_quantity = None
        elif isinstance(absolute_quantity, ContinuousValue):
            if isinstance(absolute_quantity, NominalReal) and absolute_quantity.nominal < 0:
                raise ValueError("absolute_quantity must be positive")
            elif isinstance(absolute_quantity, NormalReal) and absolute_quantity.mean < 0:
                raise ValueError("absolute_quantity must be positive")
            elif isinstance(absolute_quantity, UniformReal) and absolute_quantity.lower_bound < 0:
                raise ValueError("absolute_quantity must be positive")
            self._absolute_quantity = absolute_quantity
        else:
            raise TypeError("absolute_quantity was not given as a continuous value")



"""Bake a cake."""
import json

from taurus.client.json_encoder import thin_dumps, dumps
from taurus.entity.attribute.condition import Condition
from taurus.entity.attribute.parameter import Parameter
from taurus.entity.attribute.property import Property
from taurus.entity.bounds.integer_bounds import IntegerBounds
from taurus.entity.bounds.real_bounds import RealBounds
from taurus.entity.link_by_uid import LinkByUID
from taurus.entity.object.ingredient_run import IngredientRun
from taurus.entity.object.ingredient_spec import IngredientSpec
from taurus.entity.object.material_run import MaterialRun
from taurus.entity.object.material_spec import MaterialSpec
from taurus.entity.object.measurement_run import MeasurementRun
from taurus.entity.object.measurement_spec import MeasurementSpec
from taurus.entity.object.process_run import ProcessRun
from taurus.entity.object.process_spec import ProcessSpec
from taurus.entity.template.condition_template import ConditionTemplate
from taurus.entity.template.material_template import MaterialTemplate
from taurus.entity.template.measurement_template import MeasurementTemplate
from taurus.entity.template.parameter_template import ParameterTemplate
from taurus.entity.template.process_template import ProcessTemplate
from taurus.entity.template.property_template import PropertyTemplate
from taurus.entity.value.nominal_integer import NominalInteger
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.value.normal_real import NormalReal
from taurus.enumeration.origin import Origin
from taurus.util.impl import set_uuids


def make_cake():
    """Define all objects that go into making a demo cake."""
    cake = MaterialRun(name='Cake')

    baked = MaterialRun(name='Baked Cake')
    frosting = MaterialRun(name='Frosting')

    batter = MaterialRun(name='Batter')

    wetmix = MaterialRun(name='Wet Mix')
    drymix = MaterialRun(name='Dry Mix')

    flour = MaterialRun(name='Flour')
    baking_powder = MaterialRun(name='Baking Powder')
    salt = MaterialRun(name='Salt')
    sugar = MaterialRun(name='Sugar')
    butter = MaterialRun(name='Butter')
    eggs = MaterialRun(name='Eggs')
    vanilla = MaterialRun(name='Vanilla')
    milk = MaterialRun(name='Milk')
    chocolate = MaterialRun(name='Chocolate')
    powder_sugar = MaterialRun(name='Powdered Sugar')

    cake.process = ProcessRun(name='Icing')
    for ingredient in (baked, frosting):
        IngredientRun(material=ingredient, process=cake.process)

    cake_taste = MeasurementRun(name='Final Taste', material=cake)
    cake_appearance = MeasurementRun(name='Final Appearance', material=cake)

    baked.process = ProcessRun(name='Baking')
    IngredientRun(material=batter, process=baked.process)

    batter.process = ProcessRun(name='Mixing Batter')
    for ingredient in (wetmix, drymix, milk):
        IngredientRun(material=ingredient, process=batter.process)

    wetmix.process = ProcessRun(name='Mixing Wet')
    for ingredient in (sugar, butter, eggs, vanilla):
        IngredientRun(material=ingredient, process=wetmix.process)

    drymix.process = ProcessRun(name='Mixing Dry')
    for ingredient in (flour, baking_powder, salt):
        IngredientRun(material=ingredient, process=drymix.process)

    set_uuids(cake)

    frosting.process = ProcessRun(name='Mixing Frosting')
    IngredientRun(material=LinkByUID.from_entity(butter), process=frosting.process)
    IngredientRun(material=chocolate, process=frosting.process)
    IngredientRun(material=powder_sugar, process=frosting.process)
    IngredientRun(material=LinkByUID.from_entity(vanilla), process=frosting.process)
    IngredientRun(material=LinkByUID.from_entity(milk), process=frosting.process)

    frosting_taste = MeasurementRun(name='Frosting Taste', material=frosting)
    frosting_sweetness = MeasurementRun(name='Frosting Sweetness', material=frosting)

    ######################################################################
    # And now let's create abstract cake
    cake.spec = MaterialSpec(name='Abstract Cake')

    baked.spec = MaterialSpec(name='Abstract Baked Cake')
    frosting.spec = MaterialSpec(name='Abstract Frosting')

    batter.spec = MaterialSpec(name='Abstract Batter')

    wetmix.spec = MaterialSpec(name='Abstract Wet Mix')
    drymix.spec = MaterialSpec(name='Abstract Dry Mix')

    flour.spec = MaterialSpec(name='Abstract Flour')
    baking_powder.spec = MaterialSpec(name='Abstract Baking Powder')
    salt.spec = MaterialSpec(name='Abstract Salt')
    sugar.spec = MaterialSpec(name='Abstract Sugar')
    butter.spec = MaterialSpec(name='Abstract Butter')
    eggs.spec = MaterialSpec(name='Abstract Eggs')
    vanilla.spec = MaterialSpec(name='Abstract Vanilla')
    milk.spec = MaterialSpec(name='Abstract Milk')
    chocolate.spec = MaterialSpec(name='Abstract Chocolate')
    powder_sugar.spec = MaterialSpec(name='Abstract Powdered Sugar')

    # with abstract processes
    cake.spec.process = ProcessSpec(name='Icing, in General')
    cake.process.spec = cake.spec.process
    baked.spec.process = ProcessSpec(name='Baking, in General')
    baked.process.spec = baked.spec.process
    batter.spec.process = ProcessSpec(name='Mixing Batter, in General')
    batter.process.spec = batter.spec.process
    wetmix.spec.process = ProcessSpec(name='Mixing Wet, in General')
    wetmix.process.spec = wetmix.spec.process
    drymix.spec.process = ProcessSpec(name='Mixing Dry, in General')
    drymix.process.spec = drymix.spec.process
    frosting.spec.process = ProcessSpec(name='Mixing Frosting, in General')
    frosting.process.spec = frosting.spec.process

    # and abstract ingredients
    for run in cake.process.ingredients:
        run.spec = IngredientSpec(material=run.material.spec, process=cake.process.spec)

    for run in baked.process.ingredients:
        run.spec = IngredientSpec(material=run.material.spec, process=baked.process.spec)

    for run in batter.process.ingredients:
        run.spec = IngredientSpec(material=run.material.spec, process=batter.process.spec)

    for run in wetmix.process.ingredients:
        run.spec = IngredientSpec(material=run.material.spec, process=wetmix.process.spec)

    for run in drymix.process.ingredients:
        run.spec = IngredientSpec(material=run.material.spec, process=drymix.process.spec)

    set_uuids(cake)

    IngredientSpec(material=LinkByUID.from_entity(butter.spec), process=frosting.spec.process)
    IngredientSpec(material=chocolate.spec, process=frosting.spec.process)
    IngredientSpec(material=powder_sugar.spec, process=frosting.spec.process)
    IngredientSpec(material=LinkByUID.from_entity(vanilla.spec), process=frosting.spec.process)
    IngredientSpec(material=LinkByUID.from_entity(milk.spec), process=frosting.spec.process)

    # and spec out the measurements
    cake_taste.spec = MeasurementSpec(name='Taste')
    cake_appearance.spec = MeasurementSpec(name='Appearance')
    frosting_taste.spec = cake_taste.spec  # Taste
    frosting_sweetness.spec = MeasurementSpec(name='Sweetness')

    ######################################################################
    # Let's add some attributes
    cook_time_template = ConditionTemplate(
        name="Cooking time",
        bounds=RealBounds(0, 7 * 24.0, "hr")
    )
    oven_temperature_setting_template = ParameterTemplate(
        name="Oven temperature setting",
        bounds=RealBounds(0, 10000.0, "K")
    )
    oven_temperature_template = ConditionTemplate(
        name="Oven temperature",
        bounds=RealBounds(0, 10000.0, "K")
    )
    baking_template = ProcessTemplate(
        name="Baking in an oven",
        conditions=[(oven_temperature_template, RealBounds(0, 700, "degF"))],
        parameters=[(oven_temperature_setting_template, RealBounds(100, 550, "degF"))]
    )

    tastiness_template = PropertyTemplate(
        name="Tastiness",
        bounds=IntegerBounds(lower_bound=1, upper_bound=10)
    )
    taste_test_template = MeasurementTemplate(
        properties=[tastiness_template]
    )

    baked.process.conditions.append(Condition(name='Cooking time',
                                              template=cook_time_template,
                                              origin=Origin.MEASURED,
                                              value=NominalReal(nominal=48, units='min')))
    baked.spec.process.conditions.append(Condition(name='Cooking time',
                                                   template=cook_time_template,
                                                   origin=Origin.SPECIFIED,
                                                   value=NormalReal(mean=50, std=5, units='min')))
    baked.process.conditions.append(Condition(name='Oven temperature',
                                              origin="measured",
                                              value=NominalReal(nominal=362, units='degF')))
    baked.spec.process.parameters.append(Parameter(name='Oven temperature setting',
                                                   template=oven_temperature_setting_template,
                                                   origin="specified",
                                                   value=NominalReal(nominal=350, units='degF')))
    cake_taste.properties.append(Property(name='Tastiness',
                                          origin=Origin.MEASURED,
                                          template=tastiness_template,
                                          value=NominalInteger(nominal=5)))
    cake_appearance.properties.append(Property(name='Visual Appeal',
                                               origin=Origin.MEASURED,
                                               value=NominalInteger(nominal=5)))
    frosting_taste.properties.append(Property(name='Tastiness',
                                              origin=Origin.MEASURED,
                                              template=tastiness_template,
                                              value=NominalInteger(nominal=4)))
    frosting_sweetness.properties.append(Property(name='Sweetness (Sucrose-basis)',
                                                  origin=Origin.MEASURED,
                                                  value=NominalReal(nominal=1.7, units='')))

    baked.process.spec.template = baking_template
    cake_taste.spec.template = taste_test_template
    frosting_taste.spec.template = taste_test_template

    dessert_template = MaterialTemplate(
        name="Dessert",
        properties=[tastiness_template]
    )

    cake.spec.template = dessert_template
    frosting.spec.template = dessert_template

    return cake


if __name__ == "__main__":
    cake = make_cake()

    with open("example_taurus_material_template.json", "w") as f:
        f.write(thin_dumps(cake.template, indent=2))

    with open("example_taurus_process_template.json", "w") as f:
        f.write(thin_dumps(cake.process.ingredients[0].material.process.template, indent=2))

    with open("example_taurus_measurement_template.json", "w") as f:
        f.write(thin_dumps(cake.measurements[0].template, indent=2))

    with open("example_taurus_material_spec.json", "w") as f:
        f.write(thin_dumps(cake.spec, indent=2))

    with open("example_taurus_process_spec.json", "w") as f:
        f.write(thin_dumps(cake.process.spec, indent=2))

    with open("example_taurus_ingredient_spec.json", "w") as f:
        f.write(thin_dumps(cake.process.spec.ingredients[0], indent=2))

    with open("example_taurus_measurement_spec.json", "w") as f:
        f.write(thin_dumps(cake.measurements[0].spec, indent=2))

    with open("example_taurus_material_run.json", "w") as f:
        f.write(thin_dumps(cake, indent=2))

    with open("example_taurus_process_run.json", "w") as f:
        f.write(thin_dumps(cake.process, indent=2))

    with open("example_taurus_ingredient_run.json", "w") as f:
        f.write(thin_dumps(cake.process.ingredients[0], indent=2))

    with open("example_taurus_measurement_run.json", "w") as f:
        f.write(thin_dumps(cake.measurements[0], indent=2))

    with open("example_taurus_material_history.json", "w") as f:
        context = json.loads(dumps(cake))[0]
        f.write(json.dumps(context, indent=2))



"""Bounds an integer to be between two values."""
from taurus.entity.bounds.base_bounds import BaseBounds
from taurus.entity.value.base_value import BaseValue
from taurus.entity.value.integer_value import IntegerValue
from taurus.entity.value.nominal_integer import NominalInteger
from taurus.entity.value.uniform_integer import UniformInteger


class IntegerBounds(BaseBounds):
    """
    Bounded subset of the integers, parameterized by a lower and upper bound.

    Parameters
    ----------
    lower_bound: int
        Lower endpoint.
    upper_bound: int
        Upper endpoint.

    """

    typ = "integer_bounds"

    def __init__(self, lower_bound=None, upper_bound=None):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        if self.lower_bound is None or abs(self.lower_bound) >= float("inf"):
            raise ValueError("Lower bound must be given and finite: {}".format(self.lower_bound))

        if self.upper_bound is None or abs(self.upper_bound) >= float("inf"):
            raise ValueError("Upper bound must be given and finite")

    def validate(self, value: BaseValue) -> bool:
        """
        Checks if a value is an integer within the bounds.

        Parameters
        ----------
        value: BaseValue
            Value to validate. In order to be valid, must be an
            :py:class:`IntegerValue <taurus.entity.value.integer_value.IntegerValue>`
            and be between the lower and upper bound.

        Returns
        -------
        bool
            True if the value is between the lower and upper bound.

        """
        if not super().validate(value):
            return False
        if not isinstance(value, IntegerValue):
            return False

        if isinstance(value, NominalInteger):
            return self.upper_bound >= value.nominal >= self.lower_bound

        if isinstance(value, UniformInteger):
            return self.upper_bound >= value.upper_bound and self.lower_bound <= value.lower_bound

    def contains(self, bounds: BaseBounds) -> bool:
        """
        Check if another bounds is a subset of this range.

        The other bounds must also be an IntegerBounds and its lower and upper bound must *both*
        be within the range of this bounds object.

        Parameters
        ----------
        bounds: BaseBounds
            Other bounds object to check.

        Returns
        -------
        bool
            True if the other bounds is contained by this bounds.

        """
        if not super().contains(bounds):
            return False
        if not isinstance(bounds, IntegerBounds):
            return False

        return bounds.lower_bound >= self.lower_bound and bounds.upper_bound <= self.upper_bound



from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_quantities import HasQuantities
from taurus.entity.setters import validate_list
from taurus.entity.valid_list import ValidList


class IngredientSpec(BaseObject, HasQuantities):
    """
    An ingredient specification.

    Ingredients annotate a material with information about its usage in a process.

    Parameters
    ----------
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the ingredient spec.
    material: MaterialSpec
        Material that this ingredient is.
    process: ProcessSpec
        Process that this ingredient is used in.
    mass_fraction: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The mass fraction of the ingredient in the process.
    volume_fraction: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The volume fraction of the ingredient in the process.
    number_fraction: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The number fraction of the ingredient in the process.
    absolute_quantity: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The absolute quantity of the ingredient in the process.
    name: str, optional
        Label on the ingredient that is unique within the process that contains it.
    labels: List[str], optional
        Additional labels on the ingredient that must be unique.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    """

    typ = "ingredient_spec"

    def __init__(self, material=None, process=None, name=None, labels=None,
                 mass_fraction=None, volume_fraction=None, number_fraction=None,
                 absolute_quantity=None,
                 uids=None, tags=None, notes=None, file_links=None):
        """
        Create an IngredientSpec object.

        Assigns a unique_label and other descriptive labels to a material used as an ingredient
        :param material: MaterialSpec that is being used as the ingredient
        :param process: ProcessSpec that uses this ingredient
        :param name: of the ingredient as used in the process, i.e. "the peanut butter"
        :param labels: that this ingredient belongs to, e.g. "spread" or "solvent"
        :param mass_fraction: fraction of the ingredients that is this input ingredient, by mass
        :param volume_fraction: fraction of the ingredients that is this ingredient, by volume
        :param number_fraction: fraction of the ingredients that is this ingredient, by number
        :param absolute_quantity: quantity of this ingredient in an absolute sense, e.g. 2 cups
        """
        BaseObject.__init__(self, name, uids, tags, notes=notes, file_links=file_links)
        HasQuantities.__init__(self, mass_fraction, volume_fraction, number_fraction,
                               absolute_quantity)

        self._material = None
        self._process = None
        self._labels = None

        self.material = material
        self.process = process
        self.labels = labels

    @property
    def labels(self):
        """Get labels."""
        return self._labels

    @labels.setter
    def labels(self, labels):
        self._labels = validate_list(labels, str)

    @property
    def material(self):
        """Get the material spec."""
        return self._material

    @material.setter
    def material(self, material):
        from taurus.entity.object.material_spec import MaterialSpec
        from taurus.entity.link_by_uid import LinkByUID
        if material is None:
            self._material = None
        elif isinstance(material, (MaterialSpec, LinkByUID)):
            self._material = material
        else:
            raise ValueError("IngredientSpec.material must be a MaterialSpec")

    @property
    def process(self):
        """Get the material."""
        return self._process

    @process.setter
    def process(self, process):
        from taurus.entity.object import ProcessSpec
        from taurus.entity.link_by_uid import LinkByUID
        if process is None:
            self._process = None
        elif isinstance(process, ProcessSpec):
            self._process = process
            if not isinstance(process.ingredients, ValidList):
                process._ingredients = validate_list(self, IngredientSpec)
            else:
                process._ingredients.append(self)
        elif isinstance(process, LinkByUID):
            self._process = process
        else:
            raise ValueError("IngredientSpec.process must be a ProcessSpec")



"""A normally distributed real value."""
from taurus.entity.value.continuous_value import ContinuousValue


class NormalReal(ContinuousValue):
    """
    Normal distribution over real numbers, parameterized by a mean and standard deviation.

    Parameters
    ----------
    mean: float
        Mean of the distribution.
    std: float
        Standard deviation of the normal distribution.
    units: str
        A string describing the units. Units must be present and they must be parseable by Pint.
        An empty string can be used for the units of a dimensionless quantity.

    """

    typ = "normal_real"

    def __init__(self, mean=None, std=None, units=None):
        ContinuousValue.__init__(self, units)
        self.mean = mean
        self.std = std



"""A uniformly distributed integer value."""
from taurus.entity.value.integer_value import IntegerValue


class UniformInteger(IntegerValue):
    """
    Uniform integer distribution, with inclusive lower and upper bounds.

    Parameters
    ----------
    lower_bound: int
        Inclusive lower bound of the distribution.
    upper_bound: int
        Inclusive upper bound of the distribution.

    """

    typ = "uniform_integer"

    def __init__(self, lower_bound=None, upper_bound=None):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        assert lower_bound <= upper_bound, \
            "the lower bound must be <= the upper bound"



"""For entities that have conditions."""
from taurus.entity.attribute.condition import Condition
from taurus.entity.setters import validate_list


class HasConditions(object):
    """Mixin-trait for entities that include conditions."""

    def __init__(self, conditions):
        self._conditions = None
        self.conditions = conditions

    @property
    def conditions(self):
        """Get a list of the conditions."""
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        self._conditions = validate_list(conditions, Condition)



from taurus.entity.attribute.condition import Condition
from taurus.entity.attribute.property import Property
from taurus.entity.dict_serializable import DictSerializable
from taurus.entity.setters import validate_list


class PropertyAndConditions(DictSerializable):
    """
    A property and the conditions under which that property was determined.

    This attribute is only relevant for material specs.

    Parameters
    ----------
    property: Property
        A property attribute
    conditions: List[Condition]
        An optional list of conditions associated with this property.

    """

    typ = "property_and_conditions"

    def __init__(self, property=None, conditions=None):
        self._property = None
        self.property = property
        self._conditions = None
        self.conditions = conditions

    @property
    def conditions(self):
        """Get conditions."""
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        self._conditions = validate_list(conditions, Condition)

    # Horrible hacks to make templates work in the short term
    @property
    def name(self):
        """Get name of attribute (use name of property)."""
        return self.property.name

    @property
    def template(self):
        """Get template of attribute (use template of property)."""
        return self.property.template

    @property
    def origin(self):
        """Get origin of attribute (use origin of property)."""
        return self.property.origin

    @property
    def value(self):
        """Get value of attribute (use value of property)."""
        return self.property.value

    # NOTE: this definition must go last, or else it overrides the property decorator
    @property
    def property(self):
        """Get property."""
        return self._property

    @property.setter
    def property(self, value):
        if isinstance(value, Property):
            self._property = value
        else:
            raise ValueError("property must be a Property")



"""Base class for all enumerations."""
from enum import Enum


class BaseEnumeration(Enum):
    """Enumeration class that can convert between enumerations and associated values."""

    def __init__(self, *args):
        """Ensure that there are no duplicates in the enumeration."""
        cls = self.__class__
        if any(self.value == e.value for e in cls):
            raise ValueError("Duplicates not allowed in enumerated set of values {}".format(cls))
        if not isinstance(self.value, str):
            raise ValueError("All values of enum {} must be strings".format(cls))

    @classmethod
    def get_value(cls, name):
        """
        Return a valid value associated with name.

        If name is equal to one of the enum members, or to the value
        associated with an enum member, then return the relevant value.
        """
        if name is None:
            return None
        if any(name == e.value for e in cls):
            return name
        if any(name == e for e in cls):
            return name.value
        raise ValueError("'{}' is not a valid choice for enumeration {}".format(name, cls))

    @classmethod
    def get_enum(cls, name):
        """
        Return the enumeration associated with name.

        If name is equal to one of the enum members, or to the value
        associated with an enum member, then return the relevant enumeration.
        """
        if name is None:
            return None
        if any(name == e.value for e in cls):
            return next(e for e in cls if e.value == name)
        if any(name == e for e in cls):
            return name
        raise ValueError("'{}' is not a valid choice for enumeration {}".format(name, cls))



from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_conditions import HasConditions
from taurus.entity.object.has_parameters import HasParameters
from taurus.entity.object.has_source import HasSource


class ProcessRun(BaseObject, HasConditions, HasParameters, HasSource):
    """
    A process run.

    Processes transform zero or more input materials into exactly one output material.
    This includes links to conditions and parameters under which the process was performed,
    as well as soft links to the output material and each of the input ingredients.

    Parameters
    ----------
    name: str, optional
        Name of the process run.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the process run.
    conditions: List[Condition], optional
        Conditions under which this process run occurs.
    parameters: List[Parameter], optional
        Parameters of this process run.
    spec: ProcessSpec
        Spec for this process run.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.
    source: PerformedSource, optional
        Information about the person who performed the run and when.

    Attributes
    ----------
    output_material: MaterialRun
        The material run that this process run produces. The link is established by creating
        the material run and settings its `process` field to this process run.

    ingredients: List[IngredientRun]
        Ingredient runs that act as inputs to this process run. The link is established by
        creating each ingredient run and setting its `process` field to this process run.

    """

    typ = "process_run"

    skip = {"_output_material", "_ingredients"}

    def __init__(self, name=None, spec=None,
                 conditions=None, parameters=None,
                 uids=None, tags=None, notes=None, file_links=None, source=None):
        BaseObject.__init__(self, name=name, uids=uids, tags=tags, notes=notes,
                            file_links=file_links)
        HasConditions.__init__(self, conditions)
        HasParameters.__init__(self, parameters)
        HasSource.__init__(self, source)

        self._ingredients = []

        self._spec = None
        self.spec = spec
        self._output_material = None

    @property
    def output_material(self):
        """Get the output material run."""
        return self._output_material

    @property
    def ingredients(self):
        """Get the input ingredient runs."""
        return self._ingredients

    @property
    def spec(self):
        """Get the process spec."""
        return self._spec

    @spec.setter
    def spec(self, spec):
        from taurus.entity.object.process_spec import ProcessSpec
        from taurus.entity.link_by_uid import LinkByUID
        if spec is None:
            self._spec = None
        elif isinstance(spec, (ProcessSpec, LinkByUID)):
            self._spec = spec
        else:
            raise ValueError("spec must be a ProcessSpec: {}".format(spec))

    @property
    def template(self):
        """Get the template of the spec, if applicable."""
        from taurus.entity.object.process_spec import ProcessSpec
        if isinstance(self.spec, ProcessSpec):
            return self.spec.template
        else:
            return None



"""A unique id that stands in for a data object."""
import uuid

from taurus.entity.dict_serializable import DictSerializable


class LinkByUID(DictSerializable):
    """
    Link object, which replaces pointers to other entities before serialization and writing.

    Parameters
    ----------
    scope: str
        The scope of the unique identifier. Scopes are case-insensitive.
    id: str
        The unique identifier.

    """

    typ = "link_by_uid"

    def __init__(self, scope, id):
        # TODO: parse to make sure it's valid
        self.scope = scope
        self.id = id

    def __repr__(self):
        return str({"scope": self.scope, "uid": self.id})

    @classmethod
    def from_entity(cls, entity, name="auto"):
        """
        Create LinkByUID from in-memory object using id with scope 'name'.

        Parameters
        ----------
        entity: BaseEntity
            The entity to substitute with a LinkByUID
        name: str, optional
            The scope of the id.

        Returns
        -------
        LinkByUID
            A link object that references `entity` through its scope and id.

        """
        if name in entity.uids:
            scope, uid = name, entity.uids[name]
        else:
            if not entity.uids:
                entity.add_uid(name, str(uuid.uuid4()))
            scope, uid = next((s, i) for s, i in entity.uids.items())
        return LinkByUID(scope, uid)



"""All possible origins of an attribute."""
from taurus.enumeration.base_enumeration import BaseEnumeration


class Origin(BaseEnumeration):
    """Enumeration containing all possible origins for an attribute."""

    MEASURED = "measured"
    PREDICTED = "predicted"
    SUMMARY = "summary"
    SPECIFIED = "specified"
    COMPUTED = "computed"
    UNKNOWN = "unknown"



from taurus.entity.dict_serializable import DictSerializable


class PerformedSource(DictSerializable):
    """
    Information about an activity that was performed.

    Parameters
    ----------
    performed_by: str
        The name of the person who performed the activity
    performed_date: str
        When the activity was performed as

    """

    typ = "performed_source"

    def __init__(self, performed_by=None, performed_date=None):
        self._performed_by = None
        self._performed_date = None

        self.performed_by = performed_by
        self.performed_date = performed_date

    @property
    def performed_by(self):
        """Get the name of the person who performed the activity."""
        return self._performed_by

    @performed_by.setter
    def performed_by(self, value):
        if value is None:
            self._performed_by = None
        elif isinstance(value, str):
            self._performed_by = value
        else:
            raise TypeError("performed_by must be a string")

    @property
    def performed_date(self):
        """Get when the activity was performed."""
        return self._performed_date

    @performed_date.setter
    def performed_date(self, value):
        if value is None:
            self._performed_date = None
        elif isinstance(value, str):
            self._performed_date = value
        else:
            raise TypeError("performed_date must be a string")



"""Utility methods."""


def make_instance(base_spec):
    """
    Create a set of Run objects that mimic the connectivity of the passed Spec object.

    Paramters
    ---------
    base_spec: BaseObject
        A spec instance that may point to other specs.

    Returns
    -------
    BaseObject
        The run instance that is created, and may point to other runs.

    """
    seen = dict()

    def crawler(spec):
        from taurus.entity.object.measurement_spec import MeasurementSpec
        from taurus.entity.object.measurement_run import MeasurementRun
        from taurus.entity.object.material_spec import MaterialSpec
        from taurus.entity.object.material_run import MaterialRun
        from taurus.entity.object.ingredient_spec import IngredientSpec
        from taurus.entity.object.ingredient_run import IngredientRun
        from taurus.entity.object.process_spec import ProcessSpec
        from taurus.entity.object.process_run import ProcessRun

        if id(spec) in seen:
            return seen[id(spec)]

        if isinstance(spec, MeasurementSpec):
            seen[id(spec)] = MeasurementRun(
                name=spec.name,
                spec=spec
            )
        elif isinstance(spec, MaterialSpec):
            seen[id(spec)] = MaterialRun(
                name=spec.name,
                spec=spec
            )
            seen[id(spec)].process = crawler(spec.process) if spec.process else None
        elif isinstance(spec, IngredientSpec):
            seen[id(spec)] = IngredientRun(name=spec.name, spec=spec)
            seen[id(spec)].material = crawler(spec.material) if spec.material else None
        elif isinstance(spec, ProcessSpec):
            seen[id(spec)] = ProcessRun(
                name=spec.name,
                spec=spec
            )
            for x in spec.ingredients:
                crawler(x).process = seen[id(spec)]
        else:
            raise TypeError('Passed object is not a spec-like object({})'.format(type(spec)))

        # Should we assume that the same MaterialSpec in different parts of the tree
        # yields the same MaterialRun?
        return seen[id(spec)]

    return crawler(base_spec)


# Global to support array_like
_array_like = None


def array_like():
    """
    Figure out what kinds of list-like things we should be supporting for list type-checks.

    Returns
    -------
    tuple
        A tuple of supported array-like classes.

    """
    global _array_like
    if _array_like is not None:
        return _array_like
    try:
        import numpy as np
        try:
            import pandas as pd
            _array_like = (list, tuple, np.ndarray, pd.core.base.PandasObject)
        except ImportError:
            _array_like = (list, tuple, np.ndarray)
    except ImportError:
        _array_like = (list, tuple)

    return _array_like



"""Encoding and decoding taurus objects to/from json."""
import json
from json import JSONEncoder

from taurus.entity.attribute.condition import Condition
from taurus.entity.attribute.parameter import Parameter
from taurus.entity.attribute.property import Property
from taurus.entity.attribute.property_and_conditions import PropertyAndConditions
from taurus.entity.base_entity import BaseEntity
from taurus.entity.bounds.categorical_bounds import CategoricalBounds
from taurus.entity.bounds.composition_bounds import CompositionBounds
from taurus.entity.bounds.integer_bounds import IntegerBounds
from taurus.entity.bounds.real_bounds import RealBounds
from taurus.entity.dict_serializable import DictSerializable
from taurus.entity.file_link import FileLink
from taurus.entity.link_by_uid import LinkByUID
from taurus.entity.object import ProcessRun, MaterialRun, MeasurementRun
from taurus.entity.object.ingredient_run import IngredientRun
from taurus.entity.object.ingredient_spec import IngredientSpec
from taurus.entity.object.material_spec import MaterialSpec
from taurus.entity.object.measurement_spec import MeasurementSpec
from taurus.entity.object.process_spec import ProcessSpec
from taurus.entity.source.performed_source import PerformedSource
from taurus.entity.template.condition_template import ConditionTemplate
from taurus.entity.template.material_template import MaterialTemplate
from taurus.entity.template.measurement_template import MeasurementTemplate
from taurus.entity.template.parameter_template import ParameterTemplate
from taurus.entity.template.process_template import ProcessTemplate
from taurus.entity.template.property_template import PropertyTemplate
from taurus.entity.value.discrete_categorical import DiscreteCategorical
from taurus.entity.value.empirical_formula import EmpiricalFormula
from taurus.entity.value.nominal_categorical import NominalCategorical
from taurus.entity.value.nominal_composition import NominalComposition
from taurus.entity.value.nominal_integer import NominalInteger
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.value.normal_real import NormalReal
from taurus.entity.value.uniform_integer import UniformInteger
from taurus.entity.value.uniform_real import UniformReal
from taurus.enumeration.base_enumeration import BaseEnumeration
from taurus.util import flatten, substitute_links, deepcopy, set_uuids, substitute_objects


def dumps(obj, **kwargs):
    """
    Serialize a taurus object, or container of them, into a json-formatting string.

    Parameters
    ----------
    obj: DictSerializable or List[DictSerializable]
        The object(s) to serialize to a string.
    **kwargs: keyword args, optional
        Optional keyword arguments to pass to `json.dumps()`.

    Returns
    -------
    str
        A string version of the serialized objects.

    """
    # create a top level list of [flattened_objects, link-i-fied return value]
    res = [obj]
    additional = flatten(res)
    substitute_links(res)
    res.insert(0, additional)
    return json.dumps(res, cls=TaurusEncoder, sort_keys=True, **kwargs)


def loads(json_str, **kwargs):
    """
    Deserialize a json-formatted string into a taurus object.

    Parameters
    ----------
    json_str: str
        A string representing the serialized objects, such as what is produced by :func:`dumps`.
    **kwargs: keyword args, optional
        Optional keyword arguments to pass to `json.loads()`.

    Returns
    -------
    DictSerializable or List[DictSerializable]
        Deserialized versions of the objects represented by `json_str`, with links turned
        back into pointers.

    """
    # Create an index to hold the objects by their uid reference
    # so we can replace links with pointers
    index = {}
    raw = json.loads(json_str, object_hook=lambda x: _loado(x, index), **kwargs)
    # the return value is in the 2nd position.
    substitute_objects(raw, index)
    return raw[1]


def load(fp, **kwargs):
    """
    Load serialized string representation of an object from a file.

    Parameters
    ----------
    fp: file
        File to read.
    **kwargs: keyword args, optional
        Optional keyword arguments to pass to `json.loads()`.

    Returns
    -------
    DictSerializable or List[DictSerializable]
        Deserialized object(s).

    """
    return loads(fp.read(), **kwargs)


def dump(obj, fp, **kwargs):
    """
    Dump an object to a file, as a serialized string.

    Parameters
    ----------
    obj: DictSerializable or List[DictSerializable]
        Object(s) to dump
    fp: file
        File to write to.
    **kwargs: keyword args, optional
        Optional keyword arguments to pass to `json.dumps()`.

    Returns
    -------
    None

    """
    fp.write(dumps(obj, **kwargs))
    return


def thin_dumps(obj, **kwargs):
    """
    Serialize a "thin" version of an object in which pointers are replaced by links.

    Parameters
    ----------
    obj: BaseEntity
        Object to dump (must be a BaseEntity because it must have uids)
    **kwargs: keyword args, optional
        Optional keyword arguments to pass to `json.dumps()`.

    Returns
    -------
    str
        A serialized string of `obj`, with link_by_uid in place of pointers to other objects.

    """
    if not isinstance(obj, BaseEntity):
        raise ValueError("Can only dump BaseEntities, but got {}".format(type(obj)))

    set_uuids(obj)
    res = deepcopy(obj)
    substitute_links(res)
    return json.dumps(res, cls=TaurusEncoder, sort_keys=True, **kwargs)


def copy(obj):
    """
    Copy an object by dumping and then loading it.

    Parameters
    ----------
    obj: DictSerializable
        Object to copy

    Returns
    -------
    DictSerializable
        A copy of `obj`.

    """
    return loads(dumps(obj))


# build index from the class's typ member to the class itself
_clazzes = [
    MaterialTemplate, MeasurementTemplate, ProcessTemplate,
    MaterialSpec, MeasurementSpec, ProcessSpec, IngredientSpec,
    ProcessRun, MaterialRun, MeasurementRun, IngredientRun,
    Property, Condition, Parameter, PropertyAndConditions,
    PropertyTemplate, ConditionTemplate, ParameterTemplate,
    RealBounds, IntegerBounds, CategoricalBounds, CompositionBounds,
    NominalComposition, EmpiricalFormula,
    NominalReal, UniformReal, NormalReal, DiscreteCategorical, NominalCategorical,
    UniformInteger, NominalInteger,
    FileLink, PerformedSource
]
_clazz_index = {}
for clazz in _clazzes:
    _clazz_index[clazz.typ] = clazz


def _loado(d, index):
    if "type" not in d:
        return d
    typ = d.pop("type")

    if typ in _clazz_index:
        clz = _clazz_index[typ]
        obj = clz.from_dict(d)
    elif typ == LinkByUID.typ:
        obj = LinkByUID.from_dict(d)
        return obj
    else:
        raise ValueError("Unexpected base object type: {}".format(typ))

    if isinstance(obj, BaseEntity):
        for (scope, id) in obj.uids.items():
            index[(scope.lower(), id)] = obj
    return obj


class TaurusEncoder(JSONEncoder):
    """Rules for encoding taurus objects as json strings."""

    def default(self, o):
        """Default encoder implementation."""
        if isinstance(o, DictSerializable):
            return o.as_dict()
        elif isinstance(o, BaseEnumeration):
            return o.value
        else:
            return JSONEncoder.default(self, o)



from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_quantities import HasQuantities
from taurus.entity.setters import validate_list
from taurus.entity.valid_list import ValidList


class IngredientRun(BaseObject, HasQuantities):
    """
    An ingredient run.

    Ingredients annotate a material with information about its usage in a process.

    Parameters
    ----------
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the ingredient run.
    material: MaterialRun
        Material that this ingredient is.
    process: ProcessRun
        Process that this ingredient is used in.
    mass_fraction: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The mass fraction of the ingredient in the process.
    volume_fraction: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The volume fraction of the ingredient in the process.
    number_fraction: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The number fraction of the ingredient in the process.
    absolute_quantity: :py:class:`ContinuousValue \
    <taurus.entity.value.continuous_value.ContinuousValue>`, optional
        The absolute quantity of the ingredient in the process.
    name: str, optional
        Label on the ingredient that is unique within the process that contains it.
    labels: List[str], optional
        Additional labels on the ingredient that must be unique.
    spec: IngredientSpec
        The specification of which this ingredient is a realization.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    """

    typ = "ingredient_run"

    def __init__(self, material=None, process=None, name=None, labels=None,
                 mass_fraction=None, volume_fraction=None, number_fraction=None,
                 absolute_quantity=None,
                 spec=None, uids=None, tags=None, notes=None, file_links=None):
        BaseObject.__init__(self, name, uids, tags, notes=notes, file_links=file_links)
        HasQuantities.__init__(self, mass_fraction, volume_fraction, number_fraction,
                               absolute_quantity)

        self._material = None
        self._process = None
        self._spec = None
        self._labels = None

        self.material = material
        self.process = process
        self.spec = spec
        self.labels = labels

    @property
    def labels(self):
        """Get labels."""
        return self._labels

    @labels.setter
    def labels(self, labels):
        self._labels = validate_list(labels, str)

    @property
    def material(self):
        """Get the material."""
        return self._material

    @material.setter
    def material(self, material):
        from taurus.entity.object import MaterialRun
        from taurus.entity.link_by_uid import LinkByUID
        if material is None:
            self._material = None
        elif isinstance(material, (MaterialRun, LinkByUID)):
            self._material = material
        else:
            raise ValueError("IngredientRun.material must be a MaterialRun")

    @property
    def process(self):
        """Get the material."""
        return self._process

    @process.setter
    def process(self, process):
        from taurus.entity.object import ProcessRun
        from taurus.entity.link_by_uid import LinkByUID
        if process is None:
            self._process = None
        elif isinstance(process, ProcessRun):
            self._process = process
            if not isinstance(process.ingredients, ValidList):
                process._ingredients = validate_list(self, IngredientRun)
            else:
                process._ingredients.append(self)
        elif isinstance(process, LinkByUID):
            self._process = process
        else:
            raise ValueError("IngredientRun.process must be a ProcessRun")

    @property
    def spec(self):
        """Get the ingredient spec."""
        return self._spec

    @spec.setter
    def spec(self, spec):
        from taurus.entity.object.ingredient_spec import IngredientSpec
        from taurus.entity.link_by_uid import LinkByUID
        if spec is None:
            self._spec = None
        elif isinstance(spec, (IngredientSpec, LinkByUID)):
            self._spec = spec
        else:
            raise ValueError("spec must be a IngredientSpec: {}".format(spec))



"""A process template."""
from taurus.entity.object import ProcessRun
from taurus.entity.object.process_spec import ProcessSpec
from taurus.entity.setters import validate_list
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.template.has_condition_templates import HasConditionTemplates
from taurus.entity.template.has_parameter_templates import HasParameterTemplates


class ProcessTemplate(BaseTemplate, HasConditionTemplates, HasParameterTemplates):
    """
    A process template.

    Process templates are collections of condition and parameter templates that constrain the
    values of a measurement's condition and parameter attributes, and provide a common structure
    for describing similar measurements.

    Parameters
    ----------
    name: str, optional
        The name of the process template.
    description: str, optional
        Long-form description of the process template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    conditions: List[:class:`ConditionTemplate \
    <taurus.entity.template.condition_template.ConditionTemplate>`] or \
    List[:class:`ConditionTemplate <taurus.entity.template.condition_template.ConditionTemplate>`,\
     :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated conditions. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this condition.
    parameters: List[:class:`ParameterTemplate \
    <taurus.entity.template.parameter_template.ParameterTemplate>`] or \
    List[:class:`ParameterTemplate <taurus.entity.template.parameter_template.ParameterTemplate>`,\
     :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated parameters. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this parameter.

    """

    typ = "process_template"

    def __init__(self, name=None, description=None,
                 conditions=None, parameters=None,
                 allowed_names=None, allowed_labels=None,
                 uids=None, tags=None):
        BaseTemplate.__init__(self, name, description, uids, tags)
        HasConditionTemplates.__init__(self, conditions)
        HasParameterTemplates.__init__(self, parameters)

        self._allowed_names = None
        self.allowed_names = allowed_names

        self._allowed_labels = None
        self.allowed_labels = allowed_labels

    @property
    def allowed_names(self):
        """Get the allowed names."""
        return self._allowed_names

    @allowed_names.setter
    def allowed_names(self, allowed_names):
        # if none, leave as none; don't set to the empty set
        if allowed_names is None:
            self._allowed_names = allowed_names
        else:
            self._allowed_names = validate_list(allowed_names, str)

    @property
    def allowed_labels(self):
        """Get the allowed labels."""
        return self._allowed_labels

    @allowed_labels.setter
    def allowed_labels(self, allowed_labels):
        # if none, leave as none; don't set to the empty set
        if allowed_labels is None:
            self._allowed_labels = allowed_labels
        else:
            self._allowed_labels = validate_list(allowed_labels, str)

    def validate(self, process):
        """Check that a process satisfies all condition and parameter templates."""
        if not isinstance(process, (ProcessRun, ProcessSpec)):
            raise ValueError("ProcessTemplate can only be applied to Processes")

        self.validate_conditions(process)
        self.validate_parameters(process)

        # Check the name namespace
        if self.allowed_names is not None:
            unique_names = {x.name
                            for x in process.ingredients if x.name is not None}
            extra = unique_names.difference(set(self.allowed_names))
            if len(extra) > 0:
                raise ValueError("Found disallowed names: {}".format(extra))

        # Check the label namespace
        if self.allowed_labels is not None:
            labels = {label for x in process.ingredients for label in x.labels}
            extra = labels.difference(set(self.allowed_labels))
            if len(extra) > 0:
                raise ValueError("Found disallowed labels: {}".format(extra))



"""A restricted set of categories."""
from taurus.entity.bounds.base_bounds import BaseBounds
from taurus.entity.value.base_value import BaseValue
from taurus.entity.value.categorical_value import CategoricalValue
from taurus.entity.value.discrete_categorical import DiscreteCategorical
from taurus.entity.value.nominal_categorical import NominalCategorical
from taurus.entity.util import array_like


class CategoricalBounds(BaseBounds):
    """
    Categorical bounds, parameterized by a set of string-valued category labels.

    Parameters
    ----------
    categories: list, tuple, or set of strings
        A collection of the allowed categories

    """

    typ = "categorical_bounds"

    def __init__(self, categories=None):
        self._categories = None
        self.categories = categories

    @property
    def categories(self):
        """Get the set of categories."""
        return self._categories

    @categories.setter
    def categories(self, categories):
        if categories is None:
            self._categories = set()
        elif isinstance(categories, array_like()):
            self._categories = set(categories)
        elif isinstance(categories, set):
            self._categories = categories
        else:
            raise ValueError("Categories must be a list, tuple, or set")

        if not all(isinstance(x, str) for x in self.categories):
            raise ValueError("All the categories must be strings")

    def validate(self, value: BaseValue) -> bool:
        """
        Check if a value is in the set of allowed categories.

        Parameters
        ----------
        value: BaseValue
            Value to validate. In order to be valid, must be a
            :py:class:`CategoricalValue <taurus.entity.value.categorical_value.CategoricalValue>`
            and have be one of the allowed categories.

        Returns
        -------
        bool
            True if the value is one of the allowed categories.

        """
        if not super().validate(value):
            return False
        if not isinstance(value, CategoricalValue):
            return False

        if isinstance(value, DiscreteCategorical):
            return all(x in self.categories for x in value.probabilities)

        if isinstance(value, NominalCategorical):
            return value.category in self.categories

        msg = "Categorical bounds do not check every CategoricalValue subclass"  # pragma: no cover
        assert False, msg  # pragma: no cover

    def contains(self, bounds: BaseBounds) -> bool:
        """
        Check if another bounds object is contained by this bounds.

        The other bounds must also be a CategoricalBounds and its allowed categories must be a
        subset of this bounds's allowed categories.

        Parameters
        ----------
        bounds: BaseBounds
            Other bounds object to check.

        Returns
        -------
        bool
            True if the other bounds is contained by this bounds.

        """
        if not super().contains(bounds):
            return False
        if not isinstance(bounds, CategoricalBounds):
            return False

        return bounds.categories.issubset(self.categories)

    def as_dict(self):
        """
        Convert bounds to a dictionary.

        Returns
        -------
        dict
            A dictionary with "type" and "categories" (a sorted list of the categories).

        """
        return {"type": self.typ, "categories": sorted(list(self.categories))}



"""Utility functions."""
import uuid
from copy import deepcopy

from taurus.entity.base_entity import BaseEntity
from taurus.entity.dict_serializable import DictSerializable
from taurus.entity.link_by_uid import LinkByUID
from toolz import concatv


def set_uuids(obj, name="auto"):
    """
    Recursively assign a uuid to every BaseEntity that doesn't already contain a uuid.

    This ensures that all of the pointers in the object can be replaced with LinkByUID objects
    :param obj: to recursively assign uuids to
    :param name: of the uuid to assign (default: "auto")
    :return: None
    """
    def func(base_obj):
        if len(base_obj.uids) == 0:
            base_obj.add_uid(name, str(uuid.uuid4()))
        return
    recursive_foreach(obj, func)
    return


def substitute_links(obj, native_uid=None):
    """
    Recursively replace pointers to BaseEntity with LinkByUID objects.

    This prepares the object to be serialized or written to the API.
    It is the inverse of substitute_objects.
    It is an in-place operation.
    :param obj: target of the operation
    :param native_uid: preferred uid to use for creating LinkByUID objects (Default: None)
    :return: None
    """
    _recursive_substitute(obj, native_uid)
    return


def substitute_objects(obj, index):
    """
    Recursively replace LinkByUID objects with pointers to the objects with that UID in the index.

    This prepares the object to be used after being deserialized.
    It is the inverse of substitute_links.
    It is an in-place operation.
    :param obj: target of the operation
    :param index: containing the objects that the uids point to
    :return: None
    """
    visited = set()

    def substitute(thing):
        if id(thing) in visited:
            return
        visited.add(id(thing))
        if isinstance(thing, (list, tuple)):
            for i, x in enumerate(thing):
                if isinstance(x, LinkByUID) and (x.scope.lower(), x.id) in index:
                    thing[i] = index[(x.scope.lower(), x.id)]
                    substitute(thing[i])
                else:
                    substitute(x)
        elif isinstance(thing, dict):
            for k, v in thing.items():
                if isinstance(v, LinkByUID) and (v.scope.lower(), v.id) in index:
                    thing[k] = index[(v.scope.lower(), v.id)]
                    substitute(thing[k])
                else:
                    substitute(v)
            for k, v in thing.items():
                if isinstance(k, LinkByUID) and (k.scope.lower(), k.id) in index:
                    thing[index[(k.scope.lower(), k.id)]] = v
                    del thing[k]
                else:
                    substitute(k)
        elif isinstance(thing, DictSerializable):
            for k, v in vars(thing).items():
                if isinstance(thing, BaseEntity) and k in thing.skip:
                    continue
                if isinstance(v, LinkByUID) and (v.scope.lower(), v.id) in index:
                    # Use setattr() to call setter logic
                    setattr(thing, k.lstrip('_'), index[(v.scope.lower(), v.id)])
                    substitute(thing.__dict__[k])
                else:
                    substitute(v)
        return

    substitute(obj)
    return


def flatten(obj):
    """
    Flatten a BaseEntity into a list of objects connected by LinkByUID objects.

    This is a composite operation the amounts to:
      - Making sure at least one uid is set in each BaseEntity in scope
      - Getting a list of unique objects contained in the scope
      - Substituting the pointers in those objects with LinkByUID objects
    :param obj: defining the scope of the flatten
    :return: a list of BaseEntity with LinkByUIDs to any BaseEntity members
    """
    # The ids should be set in the actual object so they are consistent
    set_uuids(obj)

    # make a copy before we substitute the pointers for links
    copy = deepcopy(obj)

    # list of uids that we've seen, to avoid returning duplicates
    known_uids = set()

    def _flatten(base_obj):
        to_return = []
        # get all the uids of this object
        uids = list(base_obj.uids.items())

        # if none of the uids are known, then its a new object and we should return it
        if not any(uid in known_uids for uid in uids):
            to_return = [base_obj]

        # add all of the uids of this object into the known uid list
        for uid in uids:
            known_uids.add(uid)

        return to_return

    res = recursive_flatmap(copy, _flatten)
    [substitute_links(x) for x in res]
    return res


def recursive_foreach(obj, func, apply_first=False, seen=None):
    """
    Apply a function recursively to each BaseEntity object.

    :param obj: target of the operation
    :param func: to apply to each contained BaseEntity
    :param apply_first: whether to apply the func before applying it to members (default: false)
    :param seen: set of seen objects (default=None).  DON'T PASS THIS!!!
    :return: None
    """
    if seen is None:
        seen = set({})
    if obj.__hash__ is not None:
        if obj in seen:
            return
        else:
            seen.add(obj)

    if isinstance(obj, (list, tuple)):
        for i, x in enumerate(obj):
            if isinstance(x, BaseEntity):
                if apply_first:
                    func(x)
                    recursive_foreach(x, func, apply_first, seen)
                else:
                    recursive_foreach(x, func, apply_first, seen)
                    func(x)
            else:
                recursive_foreach(x, func, apply_first, seen)
    elif isinstance(obj, dict):
        for x in concatv(obj.keys(), obj.values()):
            if isinstance(x, BaseEntity):
                if apply_first:
                    func(x)
                    recursive_foreach(x, func, apply_first, seen)
                else:
                    recursive_foreach(x, func, apply_first, seen)
                    func(x)
            else:
                recursive_foreach(x, func, apply_first, seen)
    elif isinstance(obj, DictSerializable):
        for k, x in obj.__dict__.items():
            if isinstance(obj, BaseEntity) and k in obj.skip:
                continue
            if isinstance(x, BaseEntity):
                if apply_first:
                    func(x)
                    recursive_foreach(x, func, apply_first, seen)
                else:
                    recursive_foreach(x, func, apply_first, seen)
                    func(x)
            else:
                recursive_foreach(x, func, apply_first, seen)
    return


def recursive_flatmap(obj, func, seen=None):
    """
    Recursively apply and accumulate a list-valued function to BaseEntity members.

    :param obj: target of the operation
    :param func: function to apply; must be list-valued
    :param seen: set of seen objects (default=None).  DON'T PASS THIS
    :return: a list of accumulated return values
    """
    res = []

    if seen is None:
        seen = set({})
    if obj.__hash__ is not None:
        if obj in seen:
            return res
        else:
            seen.add(obj)

    if isinstance(obj, (list, tuple)):
        for i, x in enumerate(obj):
            if isinstance(x, BaseEntity):
                res.extend(recursive_flatmap(x, func, seen))
                res.extend(func(x))
            else:
                res.extend(recursive_flatmap(x, func, seen))
    elif isinstance(obj, dict):
        for x in concatv(obj.keys(), obj.values()):
            if isinstance(x, BaseEntity):
                res.extend(recursive_flatmap(x, func, seen))
                res.extend(func(x))
            else:
                res.extend(recursive_flatmap(x, func, seen))
    elif isinstance(obj, DictSerializable):
        for k, x in sorted(obj.__dict__.items()):
            if isinstance(obj, BaseEntity) and k in obj.skip:
                continue
            if isinstance(x, BaseEntity):
                res.extend(recursive_flatmap(x, func, seen))
                res.extend(func(x))
            else:
                res.extend(recursive_flatmap(x, func, seen))
    return res


def _recursive_substitute(obj, native_uid=None, seen=None):
    if seen is None:
        seen = set({})
    if obj.__hash__ is not None:
        if obj in seen:
            return
        else:
            seen.add(obj)

    if isinstance(obj, (list, tuple)):
        for i, x in enumerate(obj):
            if isinstance(x, BaseEntity):
                if len(x.uids) == 0:
                    raise ValueError("No UID for {}".format(x))
                elif native_uid and native_uid in x.uids:
                    obj[i] = LinkByUID(native_uid, x.uids[native_uid])
                else:
                    uid_to_use = next(iter(x.uids.items()))
                    obj[i] = LinkByUID(uid_to_use[0], uid_to_use[1])
            else:
                _recursive_substitute(x, native_uid, seen)
    elif isinstance(obj, dict):
        for k, x in obj.items():
            if isinstance(x, BaseEntity):
                if len(x.uids) == 0:
                    raise ValueError("No UID for {}".format(x))
                elif native_uid and native_uid in x.uids:
                    obj[k] = LinkByUID(native_uid, x.uids[native_uid])
                else:
                    uid_to_use = next(iter(x.uids.items()))
                    obj[k] = LinkByUID(uid_to_use[0], uid_to_use[1])
            else:
                _recursive_substitute(x, native_uid, seen)
        for k, x in obj.items():
            if isinstance(k, BaseEntity):
                if len(k.uids) == 0:
                    raise ValueError("No UID for {}".format(k))
                elif native_uid and native_uid in k.uids:
                    obj[LinkByUID(native_uid, k.uids[native_uid])] = x
                    del obj[k]
                else:
                    uid_to_use = next(iter(k.uids.items()))
                    obj[LinkByUID(uid_to_use[0], uid_to_use[1])] = x
                    del obj[k]
            else:
                _recursive_substitute(k, native_uid, seen)
    elif isinstance(obj, DictSerializable):
        for k, x in obj.__dict__.items():
            if isinstance(obj, BaseEntity) and k in obj.skip:
                continue
            if isinstance(x, BaseEntity):
                if len(x.uids) == 0:
                    raise ValueError("No UID for {}".format(x))
                elif native_uid and native_uid in x.uids:
                    obj.__dict__[k] = LinkByUID(native_uid, x.uids[native_uid])
                else:
                    uid_to_use = next(iter(x.uids.items()))
                    obj.__dict__[k] = LinkByUID(uid_to_use[0], uid_to_use[1])
            else:
                _recursive_substitute(x, native_uid, seen)
    return



"""All possible attributes that an entity can have."""
from taurus.enumeration.base_enumeration import BaseEnumeration


class AttributeType(BaseEnumeration):
    """Enumeration containing all possible attribute types."""

    PROPERTY = "property"
    PARAMETER = "parameter"
    CONDITION = "condition"
    METADATA = "metadata"



from taurus.entity.object.base_object import BaseObject
from taurus.entity.object.has_parameters import HasParameters
from taurus.entity.object.has_conditions import HasConditions
from taurus.entity.object.has_template import HasTemplate


class ProcessSpec(BaseObject, HasParameters, HasConditions, HasTemplate):
    """
    A process specification.

    Processes transform zero or more input materials into exactly one output material.
    This includes links to the parameters and conditions under which the process is expected
    to be performed, as well as soft links to the output material and the input ingredients.

    Parameters
    ----------
    name: str, optional
        Name of the process spec.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the process spec.
    conditions: List[Condition], optional
        Conditions under which this process spec occurs.
    parameters: List[Parameter], optional
        Parameters of this process spec.
    template: ProcessTemplate, optional
        A template bounding the valid values for this process's parameters and conditions.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    Attributes
    ----------
    output_material: MaterialSpec
        The material spec that this process spec produces. The link is established by creating
        the material spec and settings its `process` field to this process spec.

    ingredients: List[IngredientSpec], optional
        Ingredient specs that act as inputs to this process spec. The link is established by
        creating each ingredient spec and setting its `process` field to this process spec.

    """

    typ = "process_spec"

    skip = {"_output_material", "_ingredients"}

    def __init__(self, name=None, template=None,
                 parameters=None, conditions=None,
                 uids=None, tags=None, notes=None, file_links=None):
        BaseObject.__init__(self, name=name, uids=uids, tags=tags, notes=notes,
                            file_links=file_links)
        HasParameters.__init__(self, parameters=parameters)
        HasConditions.__init__(self, conditions=conditions)

        self._ingredients = []

        # By default, a ProcessSpec is not linked to any MaterialSpec.
        # If a MaterialSpec is linked to this ProcessSpec,
        # then the field self._output_material will be automatically populated
        self._output_material = None

        HasTemplate.__init__(self, template=template)

    @property
    def ingredients(self):
        """Get the list of input ingredient specs."""
        return self._ingredients

    @property
    def output_material(self):
        """Get the output material spec."""
        return self._output_material



"""A nominal integer value."""
from taurus.entity.value.integer_value import IntegerValue


class NominalInteger(IntegerValue):
    """
    Nominal integer, which does not specify an uncertainty but is not assumed to be exact.

    Parameters
    ----------
    nominal: int
        A nominal value--not assumed to be exact.

    """

    typ = "nominal_integer"

    def __init__(self, nominal=None):
        assert isinstance(nominal, int), \
            "nominal value must be an int"
        self.nominal = nominal



"""Base class for categorical values."""
from taurus.entity.value.base_value import BaseValue


class CategoricalValue(BaseValue):
    """
    Base class for categorical values, which are distributions over valid category names.

    All category names must be in unicode.
    """



"""Bound a real number to be between two values."""
from taurus.entity.bounds.base_bounds import BaseBounds
from taurus.entity.value.base_value import BaseValue
from taurus.entity.value.continuous_value import ContinuousValue
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.value.normal_real import NormalReal
from taurus.entity.value.uniform_real import UniformReal
import taurus.units as units


class RealBounds(BaseBounds):
    """
    Bounded subset of the real numbers, parameterized by a lower and upper bound.

    Parameters
    ----------
    lower_bound: float
        Lower endpoint.
    upper_bound: float
        Upper endpoint.
    default_units: str
        A string describing the units. Units must be present and they must be parseable by Pint.
        An empty string can be used for the units of a dimensionless quantity.

    """

    typ = "real_bounds"

    def __init__(self, lower_bound=None, upper_bound=None, default_units=None):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        self._default_units = None
        self.default_units = default_units

        if self.lower_bound is None or abs(self.lower_bound) >= float("inf"):
            raise ValueError("Lower bound must be given and finite: {}".format(self.lower_bound))

        if self.upper_bound is None or abs(self.upper_bound) >= float("inf"):
            raise ValueError("Upper bound must be given and finite")

    @property
    def default_units(self):
        """Get default units."""
        return self._default_units

    @default_units.setter
    def default_units(self, default_units):
        if default_units is None:
            raise ValueError("Real bounds must have units. "
                             "Use an empty string for a dimensionless quantity.")
        self._default_units = units.parse_units(default_units)

    def validate(self, value: BaseValue) -> bool:
        """
        Checks if a value is a real number within the bounds.

        Parameters
        ----------
        value: BaseValue
            Value to validate. In order to be valid, must be an
            :py:class:`ContinuousValue <taurus.entity.value.continuous_value.ContinuousValue>`
            and be between the lower and upper bound.

        Returns
        -------
        bool
            True if the value is between the lower and upper bound.

        """
        if not super().validate(value):
            return False
        if not isinstance(value, ContinuousValue):
            return False

        lower, upper = self._convert_bounds(value.units)

        if lower is None:
            return False

        if isinstance(value, NominalReal):
            return upper >= value.nominal >= lower

        # Normal distributions are interpreted as truncated normals,
        # so long as their median value is within the bounds
        if isinstance(value, NormalReal):
            return upper >= value.mean >= lower

        if isinstance(value, UniformReal):
            return upper >= value.upper_bound and lower <= value.lower_bound

    def contains(self, bounds: BaseBounds) -> bool:
        """
        Check if another bounds is a subset of this range.

        The other bounds must also be a RealBounds and its lower and upper bound must *both*
        be within the range of this bounds object.

        Parameters
        ----------
        bounds: BaseBounds
            Other bounds object to check.

        Returns
        -------
        bool
            True if the other bounds is contained by this bounds.

        """
        if not super().contains(bounds):
            return False
        if not isinstance(bounds, RealBounds):
            return False

        lower, upper = self._convert_bounds(bounds.default_units)
        if lower is None:
            return False

        return bounds.lower_bound >= lower and bounds.upper_bound <= upper

    def _convert_bounds(self, target_units):
        """
        Convert the bounds to the target unit system, or None if not possible.

        Parameters
        ----------
        target_units: str
            The units to convert into.

        Returns
        -------
        tuple (float, float)
            A tuple of the (lower_bound, upper_bound) in the target units.

        """
        # if neither have units, then just return the bounds
        if not self.default_units and not target_units:
            return self.lower_bound, self.upper_bound
        # if either have units but both don't, then we can't return anything
        elif not self.default_units or not target_units:
            return None, None
        # if both have units, then we can try to convert
        else:
            try:
                lower_bound = units.convert_units(
                    self.lower_bound, self.default_units, target_units)
                upper_bound = units.convert_units(
                    self.upper_bound, self.default_units, target_units)
                return lower_bound, upper_bound
            except units.IncompatibleUnitsError:
                return None, None



"""Composition of a material."""
from abc import abstractmethod

from taurus.entity.value.base_value import BaseValue


class CompositionValue(BaseValue):
    """Base class for composition values."""

    @property
    @abstractmethod
    def components(self):
        """Get the components in the composition."""
        ...



"""A measurement template."""
from taurus.entity.object import MeasurementRun
from taurus.entity.object.measurement_spec import MeasurementSpec
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.template.has_condition_templates import HasConditionTemplates
from taurus.entity.template.has_parameter_templates import HasParameterTemplates
from taurus.entity.template.has_property_templates import HasPropertyTemplates


class MeasurementTemplate(BaseTemplate,
                          HasPropertyTemplates, HasConditionTemplates, HasParameterTemplates):
    """
    A measurement template.

    Measurement templates are collections of condition, parameter and property templates that
    constrain the values of a measurement's condition, parameter and property attributes, and
    provide a common structure for describing similar measurements.

    Parameters
    ----------
    name: str, optional
        The name of the measurement template.
    description: str, optional
        Long-form description of the measurement template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    conditions: List[:class:`ConditionTemplate \
    <taurus.entity.template.condition_template.ConditionTemplate>`] or \
    List[:class:`ConditionTemplate <taurus.entity.template.condition_template.ConditionTemplate>`,\
     :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated conditions. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this condition.
    parameters: List[:class:`ParameterTemplate \
    <taurus.entity.template.parameter_template.ParameterTemplate>`] or \
    List[:class:`ParameterTemplate <taurus.entity.template.parameter_template.ParameterTemplate>`,\
     :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated parameters. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this parameter.
    properties: List[:class:`PropertyTemplate \
    <taurus.entity.template.property_template.PropertyTemplate>`] or \
    List[:class:`PropertyTemplate <taurus.entity.template.property_template.PropertyTemplate>`,\
     :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated properties. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this property.

    """

    typ = "measurement_template"

    def __init__(self,
                 name=None, description=None,
                 properties=None, conditions=None, parameters=None,
                 uids=None, tags=None):
        BaseTemplate.__init__(self, name, description, uids, tags)
        HasPropertyTemplates.__init__(self, properties)
        HasConditionTemplates.__init__(self, conditions)
        HasParameterTemplates.__init__(self, parameters)

    def validate(self, measurement):
        """Check that a measurement satisfies all property, condition, and parameter templates."""
        if not isinstance(measurement, (MeasurementRun, MeasurementSpec)):
            raise ValueError("MeasurementTemplate can only be applied to Measurements")
        self.validate_properties(measurement)
        self.validate_conditions(measurement)
        self.validate_parameters(measurement)



"""Base class for all continuous values."""
from taurus.entity.value.base_value import BaseValue
from taurus.units import parse_units


class ContinuousValue(BaseValue):
    """
    A base class for values that correspond to a distribution over the real numbers.

    Parameters
    ----------
    units: str
        A string describing the units. Units must be present and they must be parseable by Pint.
        An empty string can be used for the units of a dimensionless quantity.

        Examples of acceptable units: 'm', 'meter', 'metre', 'm/s^2', 'degC', 'N/meter^2',
        'joule', 'J', 'dimensionless', ''.

        Examples of unacceptable units: 'Joule', 'JOULE'.

    """

    def __init__(self, units=None):
        self._units = None
        self.units = units

    @property
    def units(self):
        """Get the units of the value."""
        return self._units

    @units.setter
    def units(self, units):
        if units is None:
            raise ValueError("Continuous values must have units. "
                             "Use an empty string for a dimensionless quantity.")
        self._units = parse_units(units)



"""Base class for integer values."""
from taurus.entity.value.base_value import BaseValue


class IntegerValue(BaseValue):
    """A base class for values that correspond to a distribution over the integers."""



"""For entities that have a condition template."""
from taurus.entity.setters import validate_list
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.template.condition_template import ConditionTemplate


class HasConditionTemplates(object):
    """
    Mixin-trait for entities that include condition templates.

    Parameters
    ----------
    conditions: List[ConditionTemplate]
        A list of this entity's condition templates.

    """

    def __init__(self, conditions):
        self._conditions = None
        self.conditions = conditions

    @property
    def conditions(self):
        """
        Get the list of condition templates.

        Returns
        -------
        List[ConditionTemplate]
            List of this entity's condition templates

        """
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        lst = validate_list(conditions, (ConditionTemplate, list, tuple))

        # make sure attribute can be a Condition
        # TODO: list.map(_.validate_scope(AttributeType.CONDITION)) all true

        self._conditions = list(map(BaseTemplate._homogenize_ranges, lst))

    def validate_conditions(self, obj):
        """Validate an object's conditions against its template."""
        BaseTemplate._validate_attributes(self, obj, "conditions")



"""Bounds a composition to have a specified set of components."""
from taurus.entity.bounds.base_bounds import BaseBounds
from taurus.entity.value.base_value import BaseValue
from taurus.entity.value.composition_value import CompositionValue
from taurus.entity.util import array_like


class CompositionBounds(BaseBounds):
    """
    Composition bounds, parameterized by a set of string-valued category labels.

    Parameters
    ----------
    components: list, tuple, or set or strings
        A collection of the components that must be present in the composition.

    """

    typ = "composition_bounds"

    def __init__(self, components=None):
        self._components = None
        self.components = components

    @property
    def components(self):
        """Get the allowed components."""
        return self._components

    @components.setter
    def components(self, value):
        if value is None:
            self._components = set()
        elif isinstance(value, array_like()):
            self._components = set(value)
        elif isinstance(value, set):
            self._components = value
        else:
            raise ValueError("Components must be a list, tuple, or set: {}".format(value))

        if not all(isinstance(x, str) for x in self.components):
            raise ValueError("All the components must be strings")

    def validate(self, value: BaseValue) -> bool:
        """
        Check if a value has all of the required components.

        Parameters
        ----------
        value: BaseValue
            Value to validate. In order to be valid, must be a
            :py:class:`CompositionValue <taurus.entity.value.composition_value.CompositionValue>`
            and have all of the required components.

        Returns
        -------
        bool
            True if the value is a composition with all required components.

        """
        if not super().validate(value):
            return False
        if not isinstance(value, CompositionValue):
            return False

        return all(x in self.components for x in value.components)

    def contains(self, bounds: BaseBounds) -> bool:
        """
        Check if another bounds is contained by this bounds.

        The other bounds must also be a CompositionBounds and its components must be a subset of
        this bounds's set of required components.

        Parameters
        ----------
        bounds: BaseBounds
            Other bounds object to check.

        Returns
        -------
        bool
            True if the other bounds is contained by this bounds.

        """
        if not super().contains(bounds):
            return False
        if not isinstance(bounds, CompositionBounds):
            return False

        return bounds.components.issubset(self.components)

    def as_dict(self):
        """
        Convert bounds to a dictionary.

        Returns
        -------
        dict
            A dictionary with "type" and "components" (a sorted list of the components).

        """
        return {"type": self.typ, "components": sorted(list(self.components))}



class CaseInsensitiveDict(dict):
    """
    A dictionary in which the keys are case-insensitive.

    It is initialized the same way as a typical dict, but the values can be accessed without
    regard to key case. The value associated with key "Key" can also be accessed with "key"
    or "KEY" or "kEy".

    Parameters
    ----------
    seq: iterable or mapping, optional
        The key-value pairs of the dictionary. Can either be a mapping object with (key, value)
        pairs, or an iterable of tuples of the form (key, value).
    **kwargs: keyword args, optional
        An alternative way of initializing the dictionary with key-value pairs.
        Example: CaseInsensitiveDict(one=1, two="two").

    """

    def __init__(self, seq=None, **kwargs):
        super().__init__(seq or {}, **kwargs)
        self.lowercase_dict = {}
        for key in self:
            self._register_key(key)

    def __getitem__(self, key: str):
        return super().__getitem__(self.lowercase_dict[key.lower()])

    def get(self, key: str):
        """
        Get the value for a given case-insensitive key.

        Parameters
        ----------
        key: str
            The key to look up (possibly with a different casing).

        Returns
        -------
        Any
            The value associated with the case-insensitive version of `key`, or None
            if `key` is not present.

        """
        return self.__getitem__(key)

    def __setitem__(self, key: str, value):
        self._register_key(key)
        super().__setitem__(key, value)

    def __contains__(self, key: str):
        return self.lowercase_dict.__contains__(key.lower())

    def _register_key(self, key: str):
        """
        Register a key to the dictionary.

        Check to make sure it doesn't already exist in a different case.

        Parameters
        ----------
        key: str
            The key to register.

        """
        prev = self.lowercase_dict.get(key.lower())
        if prev is not None and prev != key:
            raise ValueError(
                "Key '{}' already exists in dict with different case: '{}'".format(key, prev))
        self.lowercase_dict[key.lower()] = key



"""All possible types of samples."""
from taurus.enumeration.base_enumeration import BaseEnumeration


class SampleType(BaseEnumeration):
    """Enumeration containing all possible sample types for a MaterialRun."""

    EXPERIMENTAL = "experimental"
    PRODUCTION = "production"
    VIRTUAL = "virtual"
    UNKNOWN = "unknown"



"""Ingest a table."""
from taurus.entity.attribute.condition import Condition
from taurus.entity.attribute.property import Property
from taurus.entity.object import MeasurementRun
from taurus.entity.value.nominal_real import NominalReal

known_properties = ["vapor pressure"]
known_conditions = ["temperature"]


def ingest_table(material_run, table):
    """Ingest a material run into an existing table."""
    for _, row in table.iterrows():
        exp = MeasurementRun()
        for prop_name in known_properties:
            if prop_name in row:
                exp.properties.append(Property(name=prop_name,
                                               value=NominalReal(row[prop_name], '')))
        for cond_name in known_conditions:
            if cond_name in row:
                exp.conditions.append(Condition(name=cond_name,
                                                value=NominalReal(row[cond_name], '')))
        exp.material = material_run

    return material_run



from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop
import sys
import subprocess


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        """Run strip-hints to support python 3.5.

        Python version is constrained to [3.5, 4.0), so if the minor version is < 6,
        run the script to strip type hints
        """
        if sys.version_info.minor < 6:
            subprocess.call('chmod 755 scripts/strip_hints.sh', shell=True)
            subprocess.call('./scripts/strip_hints.sh', shell=True)
        install.run(self)


class PostDevelopCommand(develop):
    """Post-installation for develop mode."""

    def run(self):
        """Run strip-hints to support python 3.5.

        Python version is constrained to [3.5, 4.0), so if the minor version is < 6,
        run the script to strip type hints
        """
        if sys.version_info.minor < 6:
            subprocess.call('chmod 755 scripts/strip_hints.sh', shell=True)
            subprocess.call('./scripts/strip_hints.sh', shell=True)
        develop.run(self)


setup(name='taurus-citrine',
      version='0.0.0a0',
      url='http://github.com/CitrineInformatics/taurus-py',
      description='Python library for the Citrine Platform',
      author='Max Hutchinson',
      author_email='maxhutch@citrine.io',
      packages=find_packages(),
      package_data={
          'taurus': [
              'demo/strehlow_and_cook.json',
              'units/citrine_en.txt'
          ]
      },
      install_requires=[
          "toolz",
          "pytest>=4.3",
          "enum34",
          "pint>=0.9",
          "pymatgen",
          "strip-hints>=0.1.5"
      ],
      cmdclass={
          'install': PostInstallCommand,
          'develop': PostDevelopCommand
      })



"""Base class for all values."""
from taurus.entity.dict_serializable import DictSerializable


class BaseValue(DictSerializable):
    """
    Base class for all values.

    "Value" is a generic term for the information contained in an
    :class:`attribute <taurus.entity.attribute.base_attribute.BaseAttribute>`.
    A value may be one of the following types: `RealValue`, `IntegerValue`, `Categorical`.
    """

    typ = "value"

    def __repr__(self):
        return str(self.as_dict())

    def __eq__(self, other):
        if isinstance(other, BaseValue):
            return self.__repr__() == other.__repr__()
        else:
            return False



"""For entities that have a parameter template."""
from taurus.entity.setters import validate_list
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.template.parameter_template import ParameterTemplate


class HasParameterTemplates(object):
    """
    Mixin-trait for entities that include parameter templates.

    Parameters
    ----------
    parameters: List[ParameterTemplate]
        A list of this entity's parameter templates.

    """

    def __init__(self, parameters):
        self._parameters = None
        self.parameters = parameters

    @property
    def parameters(self):
        """
        Get the list of parameter templates.

        Returns
        -------
        List[ParameterTemplate]
            List of this entity's parameter templates

        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        lst = validate_list(parameters, (ParameterTemplate, list, tuple))

        # make sure attribute can be a Parameter
        # TODO: list.map(_.validate_scope(AttributeType.PARAMETER)) all true

        self._parameters = list(map(BaseTemplate._homogenize_ranges, lst))

    def validate_parameters(self, obj):
        """Validate an object's parameters against its templates."""
        BaseTemplate._validate_attributes(self, obj, "parameters")



"""For entities that have properties."""
from taurus.entity.attribute.property import Property
from taurus.entity.setters import validate_list


class HasProperties(object):
    """Mixin-trait for entities that include properties."""

    def __init__(self, properties):
        self._properties = None
        self.properties = properties

    @property
    def properties(self):
        """Get a list of the properties."""
        return self._properties

    @properties.setter
    def properties(self, properties):
        self._properties = validate_list(properties, Property)



"""A material template."""
from taurus.entity.object import MaterialRun
from taurus.entity.object.material_spec import MaterialSpec
from taurus.entity.template.base_template import BaseTemplate
from taurus.entity.template.has_property_templates import HasPropertyTemplates


class MaterialTemplate(BaseTemplate, HasPropertyTemplates):
    """
    A material template.

    Material templates are collections of property templates that constrain the values of
    a material's property attributes, and provide a common structure for describing similar
    materials.

    Parameters
    ----------
    name: str, optional
        The name of the material template.
    description: str, optional
        Long-form description of the material template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    properties: List[:class:`PropertyTemplate \
    <taurus.entity.template.property_template.PropertyTemplate>`] or \
    List[:class:`PropertyTemplate <taurus.entity.template.property_template.PropertyTemplate>`,\
     :py:class:`BaseBounds <taurus.entity.bounds.base_bounds.BaseBounds>`], optional
        Templates for associated properties. Each template can be provided by itself, or as a list
        with the second entry being a separate, *more restrictive* Bounds object that defines
        the limits of the value for this property.

    """

    typ = "material_template"

    def __init__(self, name=None, description=None, properties=None, uids=None, tags=None):
        BaseTemplate.__init__(self, name, description, uids, tags)
        HasPropertyTemplates.__init__(self, properties)

    def validate(self, material):
        """Check that a material satisfies all property templates."""
        if not isinstance(material, (MaterialRun, MaterialSpec)):
            raise ValueError("MaterialTemplate can only be applied to Materials")
        self.validate_properties(material)



from taurus.entity.object.base_object import BaseObject
from taurus.enumeration import SampleType


class MaterialRun(BaseObject):
    """
    A material run.

    This includes a link to the originating process and soft links to measurements.

    Parameters
    ----------
    name: str, optional
        Name of the material run.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the material run.
    process: ProcessRun
        Process that produces this material.
    sample_type: str, optional
        The form of this sample. Optionals are "experimental", "virtual", "production", or
        "unknown." Default is "unknown."
    spec: MaterialSpec
        The material specification of which this is an instance.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    Attributes
    ----------
    measurements: List[MeasurementRun], optional
        Measurements performed on this material. The link is established by creating the
        measurement run and settings its `material` field to this material run.

    """

    typ = "material_run"

    skip = {"_measurements"}

    def __init__(self, name=None, spec=None, process=None, sample_type="unknown",
                 uids=None, tags=None, notes=None, file_links=None):
        BaseObject.__init__(self, name=name, uids=uids, tags=tags, notes=notes,
                            file_links=file_links)
        self._process = None
        self._measurements = []
        self._sample_type = None
        self._spec = None

        self.spec = spec
        self.process = process
        self.sample_type = sample_type

    @property
    def process(self):
        """Get the originating process run."""
        return self._process

    @process.setter
    def process(self, process):
        from taurus.entity.object.process_run import ProcessRun
        from taurus.entity.link_by_uid import LinkByUID
        if process is None:
            self._process = None
        elif isinstance(process, LinkByUID):
            self._process = process
        elif isinstance(process, ProcessRun):
            process._output_material = self
            self._process = process
        else:
            raise ValueError("process must be a ProcessRun: {}".format(process))

    @property
    def measurements(self):
        """Get a list of measurement runs."""
        return self._measurements

    @property
    def sample_type(self):
        """Get the sample type."""
        return self._sample_type

    @sample_type.setter
    def sample_type(self, sample_type):
        self._sample_type = SampleType.get_value(sample_type)

    @property
    def spec(self):
        """Get the material spec."""
        return self._spec

    @spec.setter
    def spec(self, spec):
        from taurus.entity.object.material_spec import MaterialSpec
        from taurus.entity.link_by_uid import LinkByUID
        if spec is None:
            self._spec = None
        elif isinstance(spec, (MaterialSpec, LinkByUID)):
            self._spec = spec
        else:
            raise ValueError("spec must be a MaterialSpec: {}".format(spec))

    @property
    def template(self):
        """Get the template of the spec, if applicable."""
        from taurus.entity.object.material_spec import MaterialSpec
        if isinstance(self.spec, MaterialSpec):
            return self.spec.template
        else:
            return None



from taurus.entity.base_entity import BaseEntity
from taurus.entity.file_link import FileLink
from taurus.entity.setters import validate_list, validate_str


class BaseObject(BaseEntity):
    """
    Base class for objects.

    This includes {Material, Process, Measurement, Ingredient} {Run, Spec}

    Parameters
    ----------
    name: str, optional
        Name of the material spec.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/taurus-documentation/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/taurus-documentation/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the material spec.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    """

    def __init__(self, name=None, uids=None, tags=None, notes=None, file_links=None):
        BaseEntity.__init__(self, uids, tags)
        self.notes = notes
        self._name = None
        self._file_links = None

        self.name = name
        self.file_links = file_links

    @property
    def name(self):
        """Get name."""
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            self._name = None
        else:
            self._name = validate_str(name)

    @property
    def file_links(self):
        """Get file links."""
        return self._file_links

    @file_links.setter
    def file_links(self, file_links):
        self._file_links = validate_list(file_links, FileLink)



"""Demo representing Strehlow & Cook bandgap data with data concepts."""
from taurus.entity.util import make_instance

from taurus.entity.object.process_spec import ProcessSpec
from taurus.entity.object.material_spec import MaterialSpec
from taurus.entity.object.measurement_spec import MeasurementSpec

from taurus.entity.template.process_template import ProcessTemplate
from taurus.entity.template.material_template import MaterialTemplate
from taurus.entity.template.measurement_template import MeasurementTemplate

from taurus.entity.attribute.property import Property
from taurus.entity.template.property_template import PropertyTemplate
from taurus.entity.attribute.property_and_conditions import PropertyAndConditions

from taurus.entity.bounds.categorical_bounds import CategoricalBounds
from taurus.entity.value.nominal_categorical import NominalCategorical

from taurus.entity.bounds.composition_bounds import CompositionBounds
from taurus.entity.value.empirical_formula import EmpiricalFormula

from taurus.entity.value.normal_real import NormalReal
from taurus.entity.value.nominal_real import NominalReal
from taurus.entity.bounds.real_bounds import RealBounds


def make_strehlow_objects(table=None):
    """Make a table with Strehlow & Cook data."""
    if table is None:
        table = [["Bi$_{2}$Te$_{3}$", "Bi2Te3", 0.153, None, "Single crystalline", None],
                 ["DyN", "DyN", 2.1, None, None, None],
                 ["InAs", "InAs", 0.404, None, "Single crystalline", "Dark Gray"],
                 ["CaS", "CaS", 6.0, 0.2, "Polycrystalline", None],
                 ["Rb$_{3}$Sb", "Rb3Sb", 1.7, 0.1, None, None],
                 ["CeN", "CeN", 0.7, None, None, "Bronze"],
                 ["BaI", "BaI", None, None, None, None],
                 ["CoF$_{2}$", "CoF2", None, None, "Polycrystalline", None],
                 ["B$_{2}$Se$_{3}$", "B2Se3", None, None, None, "Orange"],
                 ["Mg$_{3}$As$_{2}$", "Mg3As2", 2.55, 0.35, "Polycrystalline", "Brown"],
                 ["Bi$_{0.85}$Sb$_{0.15}$", None, 0.01, None, "Single crystalline", None],
                 ["CoTe$_{1.88}$", None, 0.2, None, None, None]]

    cryst_bnds = CategoricalBounds(['Amorphous', 'Polycrystalline', 'Single crystalline'])
    color_bnds = CategoricalBounds(
        ['Amber', 'Black', 'Blue', 'Bluish', 'Bronze', 'Brown', 'Brown-Black', 'Copper-Red',
         'Dark Brown', 'Dark Gray', 'Dark Green', 'Dark Red', 'Gray', 'Light Gray', 'Ocher',
         'Orange', 'Orange-Red', 'Pale Yellow', 'Red', 'Red-Yellow', 'Violet', 'White',
         'Yellow', 'Yellow-Orange', 'Yellow-White'
         ])
    band_bnds = RealBounds(lower_bound=0, upper_bound=100, default_units='eV')

    chem_tmpl = PropertyTemplate(
        name="Identity",
        bounds=CompositionBounds(components=EmpiricalFormula.all_elements()),
    )
    cryst_tmpl = PropertyTemplate(
        name="Crystallinity",
        bounds=cryst_bnds,
    )
    color_tmpl = PropertyTemplate(
        name="Color",
        bounds=color_bnds,
    )
    band_tmpl = PropertyTemplate(
        name="Band gap",
        bounds=band_bnds,
    )

    # Templates
    proc_tmpl = ProcessTemplate(name='Literature source template')

    chem_mat_tmpl = MaterialTemplate(name='Chemical',
                                     properties=[chem_tmpl]
                                     )
    cryst_msr_tmpl = MeasurementTemplate(name='Crystal description',
                                         properties=[cryst_tmpl]
                                         )
    color_msr_tmpl = MeasurementTemplate(name='Color description',
                                         properties=[color_tmpl]
                                         )
    band_msr_tmpl = MeasurementTemplate(name='Band gap measurement',
                                        properties=[band_tmpl]
                                        )

    # Specs
    cryst_msr_spec = MeasurementSpec(name='Crystallinity',
                                     template=cryst_msr_tmpl
                                     )

    color_msr_spec = MeasurementSpec(name='Color',
                                     template=color_msr_tmpl
                                     )

    band_msr_spec = MeasurementSpec(name='Band gap',
                                    template=band_msr_tmpl
                                    )

    compounds = []
    max_len = 0
    for row in table:
        spec = MaterialSpec(name=row[0],
                            template=chem_mat_tmpl,
                            process=ProcessSpec(name='Literature source',
                                                template=proc_tmpl
                                                )
                            )
        run = make_instance(spec)
        compounds.append(run)
        max_len = max(max_len, len(row[0]))

        if row[1] is not None:
            spec.properties.append(
                PropertyAndConditions(property=Property(name=chem_tmpl.name,
                                                        value=EmpiricalFormula(formula=row[1]),
                                                        template=chem_tmpl)
                                      ))
        if row[2] is not None:
            band_meas = make_instance(band_msr_spec)
            band_meas.material = run
            if row[3] is not None:
                band_meas.properties.append(
                    Property(name=band_msr_tmpl.name,
                             value=NormalReal(mean=row[2],
                                              units=band_bnds.default_units,
                                              std=row[3]
                                              )
                             )
                )
            else:
                band_meas.properties.append(
                    Property(name=band_msr_tmpl.name,
                             value=NominalReal(nominal=row[2],
                                               units=band_bnds.default_units
                                               )
                             )
                )
        if row[4] is not None:
            cryst_msr = make_instance(cryst_msr_spec)
            cryst_msr.material = run
            val = NominalCategorical(category=row[4])
            cryst_msr.properties.append(Property(name=cryst_msr_tmpl.name, value=val))

        if row[5] is not None:
            color_msr = make_instance(color_msr_spec)
            color_msr.material = run
            val = NominalCategorical(category=row[5])
            color_msr.properties.append(Property(name=color_msr_tmpl.name, value=val))

    # Run through validators; these should be run automatically in theory/eventually
    # It also appears this is not actually validating
    for comp in compounds:
        if comp.spec.properties:
            if not chem_mat_tmpl.properties[0][1].validate(comp.spec.properties[0].property.value):
                raise ValueError(
                    'Material value mismatch: {}'.format(comp.spec.properties[0].value))
            chem_mat_tmpl.validate(comp.spec)
        for msr in comp.measurements:
            if not msr.template.properties[0][1].validate(msr.properties[0].value):
                raise ValueError(
                    'Measurement value mismatch: {}'.format(comp.spec.properties[0].value))
            msr.template.validate(msr)

    return compounds


def make_strehlow_table(compounds):
    """
    Headers and content for the output of make_strehlow_objects.

    Note that this is supposed to be mimicking the transformation of a set of Material Histories
    into a Training Table, and as such we are missing the column definition component of the query
    that created this particular result set.

    :param compounds: a list of MaterialRun objects from the make_strehlow_objects method
    :return:
    """
    # Stash templates in convenience variables
    for comp in compounds:
        if comp.spec.properties:
            chem_tmpl = comp.spec.properties[0].property.template
            break

    chem_mat_tmpl = compounds[0].spec.template

    for comp in compounds:
        if len(comp.measurements) == 3:  # Full list
            band_msr = comp.measurements[0].spec
            cryst_msr = comp.measurements[1].spec
            color_msr = comp.measurements[2].spec
            break

    # Consider how to specify relevant data pathing here
    output = {'headers': [], 'content': []}

    # "Chemical" is supposed to be the unifying characterization of all the root elements of the
    # Material Histories, but that can't be the spec name because the spec is Compound specific --
    # that's where the chemical is defined
    output['headers'].append(
        {'name': [chem_mat_tmpl.name,
                  "Display name"  # It would be good to derive this from the structure somehow
                  ],
         'primitive': True
         }
    )
    output['headers'].append(
        {'name': [chem_mat_tmpl.name,
                  chem_tmpl.name
                  ],
         'primitive': False,
         'bounds': CompositionBounds()
         }
    )
    output['headers'].append(
        {'name': [chem_mat_tmpl.name,
                  band_msr.name,
                  band_msr.template.properties[0][0].name
                  ],
         'primitive': False,
         'bounds': band_msr.template.properties[0][0].bounds
         }
    )
    output['headers'].append(
        {'name': [chem_mat_tmpl.name,
                  cryst_msr.name,
                  cryst_msr.template.properties[0][0].name
                  ],
         'primitive': False,
         'bounds': cryst_msr.template.properties[0][0].bounds
         }
    )
    output['headers'].append(
        {'name': [chem_mat_tmpl.name,
                  color_msr.name,
                  color_msr.template.properties[0][0].name
                  ],
         'primitive': False,
         'bounds': color_msr.template.properties[0][0].bounds
         }
    )

    for comp in compounds:
        row = [comp.spec.name]
        x = list(filter(lambda y: y.name == chem_tmpl.name, comp.spec.properties))
        if x:
            row.append(x[0].value)
        else:
            row.append(None)

        x = list(filter(lambda y: y.name == band_msr.name, comp.measurements))
        if x:
            row.append(x[0].properties[0].value)
        else:
            row.append(None)

        x = list(filter(lambda y: y.name == cryst_msr.name, comp.measurements))
        if x:
            row.append(x[0].properties[0].value)
        else:
            row.append(None)

        x = list(filter(lambda y: y.name == color_msr.name, comp.measurements))
        if x:
            row.append(x[0].properties[0].value)
        else:
            row.append(None)

        output['content'].append(row)

    return output


if __name__ == "__main__":
    import pkg_resources
    import json
    import taurus.client.json_encoder as je

    # Whether to use the full data set.
    # If `False` a minimal, predefined subset of compounds will be used.
    use_full_table = False

    if use_full_table:
        resource = pkg_resources.resource_stream("taurus.demo", "strehlow_and_cook.json")
        table = json.load(resource)
    else:
        table = None

    compounds = make_strehlow_objects(table)
    sac_tbl = make_strehlow_table(compounds)

    # Look at each different combination of Value types in a S&C record
    seen = set()
    for comp, row in zip(compounds, sac_tbl['content']):
        mask = ''.join(map(lambda x: str(type(x)), row))
        if mask not in seen:  # this is a novel shape
            seen.add(mask)
            print(comp.name)
    print('Total number of prototypes: {}'.format(len(seen)))

    print("\n\nJSON -- Training table")
    print(json.dumps(json.loads(je.dumps(sac_tbl))[1], indent=2))

    print("\n\nCSV -- Display table")
    header = list(map(lambda x: '~'.join(x['name']), sac_tbl['headers']))
    header.insert(2, header[2])
    header[0] += '~Label'
    header[1] += '~Formula'
    header[2] += '~Mean ({})'.format('eV')  # .format(sac_tbl['headers'][2]['bounds'].default_units
    header[3] += '~Std deviation ({})'.format('eV')
    header[4] += '~Category'
    header[5] += '~Category'

    print(','.join(header))

    for comp in sac_tbl['content']:
        row = [comp[0]]

        if comp[1] is None:
            row.append('')
        else:
            row.append(comp[1].formula)

        if comp[2] is None:
            row.extend(['', ''])
        elif isinstance(comp[2], NominalReal):
            row.extend([str(comp[2].nominal), ''])
        else:
            row.extend([str(comp[2].mean), str(comp[2].std)])

        if comp[3] is None:
            row.append('')
        else:
            row.append(comp[3].category)

        if comp[4] is None:
            row.append('')
        else:
            row.append(comp[4].category)

        print(','.join(row))



from taurus.entity.attribute.base_attribute import BaseAttribute
from taurus.enumeration import AttributeType


class Condition(BaseAttribute):
    """
    Condition of a property, process, or measurement.

    Conditions are environmental variables (typically measured) that may affect a process
    or measurement: e.g., temperature, pressure.

    """

    typ = "condition"
    attribute_type = AttributeType.CONDITION
